
/************************************************************************
 *
 * File: Ak.cc
 *
 * Copyright (c) 1995 by Rensselaer Polytechnic Institute,
 *                       Troy, NY 12180
 *
 * Created:       Wed Oct 11 18:51:20 1995 by James Miller
 * Last Modified: Mon Apr  8 11:27:09 1996 by James Miller
 *
 * Class to maintain the matrix of ak = E[uk] values.  This table
 * is stored as class data, with routines to build and read data values.
 *
 * E[uk] is the expected value of the kth absolute residual from a
 * Gaussian distribution with a standard deviation of 1 (and mean 0).
 *
 * An instance of the table is needed to access the values, if the
 * instance is the first created, then the static table is built,
 * otherwise nothing happens.  There is no instance data.
 *
 * Note that the table is indexed as ak[k][N].  This will return
 * the E[uk] when the population size is N.  k and N must be greater
 * than 0, and k must be less than or equal to N. Due to this indexing
 * notation, care must be taken not to overwrite the table.
 *
 ************************************************************************/

#ident "@(#)Ak.cc	1.3 96/04/08 RPI"

#include <assert.h>
#include <stream.h>
#include "Ak.h"

// initialize the necessary static data
bool Ak::built = false;

// g++ requires that static array data be initialized in order to define
// the symbol.  While the Apogee compiler does not require this, it is
// probably better C++ code to specify an initialization.
double Ak::table[MAX_N+1][MAX_N+1] = {0.0};

//
// table stored in the upper triangle
//
void Ak::build()
{
    int i,j;
    
    cerr << "Ak: Building E[uk] table." << endl;

    for (i=1; i <= MAX_N; i++)
      for (j=i; j <= MAX_N; j++)
      {
	// We do not have a public routine to calculate the inverse
	// error function.  
	// table[i][j] = M_SQRT2 * erfinv((double)i/(double)(j+1));

	// However, we do have a public routine that calculates the inverse
	// CDF of a standardized Gaussian (which is what we want anyway).
	table[i][j]=gaussian_cdf_inv(0.5* (1.0 + ((double)i / (double)(j+1))));
      }
	
    built = true;
}



// access routine
double* Ak::operator[](int row) const
{
    // make sure the row is in the table
    assert( (row > 0) && (row <= MAX_N) );

    return (table[row]);
}
