
/************************************************************************
 *
 * File: E_s_k.cc
 *
 * Copyright (c) 1995 by Rensselaer Polytechnic Institute,
 *                       Troy, NY 12180
 *
 * Created:       Wed Oct 11 19:15:11 1995 by James Miller
 * Last Modified: Mon Apr  8 11:27:40 1996 by James Miller
 *
 * Class to maintain the matrix of E[min s | k] values.  This table
 * is stored as class data, with routines to build and read data values.
 *
 * E[min s | k] is the expected value of s given that the minimum occurs at 
 * the kth absolute residual. E[min s | k] was calculated assuming a Gaussian
 * distribution with a standard deviation of 1 (and mean 0).
 *
 * An instance of the table is needed to access the values, if the
 * instance is the first created, then the static table is read,
 * otherwise nothing happens.  There is no instance data.
 *
 * Note that the table is indexed as esk[k][N].  This will return
 * the E[min s | k] when the population size is N.  k and N must be greater
 * than 0, and k must be less than or equal to N. Due to this indexing
 * notation, care must be taken not to overwrite the table.
 * 
 * This table is read from disk E_s_k.table.  The format of the file
 * is an integer indicating a value of N followed by N doubles of the
 * E[min s | k] for 1 <= k <= N.
 *
 ************************************************************************/

#ident "@(#)E_s_k.cc	1.4 96/04/08 RPI"

#include <iostream.h>
#include <fstream.h>
#include <assert.h>

#include "E_s_k.h"


// initialize the necessary static data
bool E_s_k::built = false;
int  E_s_k::maximum_N = MAX_N;

// g++ requires that static array data be initialized in order to define
// the symbol.  While the Apogee compiler does not require this, it is
// probably better C++ code to specify an initialization.
double E_s_k::table[MAX_N+1][MAX_N+1] = {0.0};


//
// table stored in the upper triangle
//
void E_s_k::build()
{
    ifstream infile("E_s_k.table", ios::in);
    int i, N;
    double val[MAX_N];

    cerr << "E_s_k: reading E[min s|k] table." << endl;
    
    if (!infile)
    {
	cerr << "E_s_k: cannot open E_s_k.table file." << endl;
	return;
    }

    while ( infile.peek() != EOF )
    {
	infile.read((char *) &N, sizeof(int) );
	if (infile.gcount() != sizeof(int))
	{
	    cerr << "E_s_k: error reading table column identifier." << endl;
	    break;
	}

	// if find an invalid N, terminate the read
	if ((N > 0) && (N <= MAX_N))
	{
	  maximum_N = N;
	  
	  infile.read((char *)val, N*sizeof(double));
	  if (infile.gcount() != N*sizeof(double))
	  {
	    cerr << "E_s_k: error reading table column "
		 << "(column = " << N << ")." << endl;
	    break;
	  }

	  // copy the data read into the table
	  for (i=1; i <= N; i++)
	    table[i][N] = val[i-1];
	}
	else
	{
	    cerr << "E_s_k: invalid column identifier, terminating read."
		 << endl;
	    break;
	}
    }
	  
    built = true;

    infile.close();
}



// access routine
double* E_s_k::operator[](int row) const
{
    // make sure the row is in the table
    assert( (row > 0) && (row <= maximum_N) );

    return (table[row]);
}


