
#ifndef _MUSEPARAMETERS_H
#define _MUSEPARAMETERS_H

/************************************************************************
 *
 * File: MuseParameters.h
 *
 * Copyright (c) 1995 by Rensselaer Polytechnic Institute,
 *                       Troy, NY 12180
 *
 * Created:       Sun Oct 15 17:53:34 1995 by James Miller
 * Last Modified: Mon Apr  8 11:24:31 1996 by James Miller
 *
 * An object to hold all the parameters (and hold the routine to parse
 * the parameter file) for the MUSE algorithm
 *
 ************************************************************************/

#ident "@(#)MuseParameters.h	1.5 96/04/08 RPI"

#ifndef bool
#include <bool.h>
#endif
#include <iostream.h>
#include <fstream.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <values.h>

#ifdef APOGEEE
extern "C" { extern int strcasecmp(const char *, const char *); };
#endif


// The xlC compiler does not have a strcasecmp() function, so we need
// to role our own
#ifdef __xlC__
#include <ctype.h>

int strcasecmp(const char *s1, const char *s2)
{
    int i, len1, len2, diff;

    len1 = strlen(s1);
    len2 = strlen(s2);
    
    for (i=0; i<MY_MIN(len1, len2); i++) {
      diff = toupper(s1[i]) - toupper(s2[i]);
      if (diff != 0) return diff;
    }

    // if strings match up to the shorter of the two strings, then the shorter
    // string is "less than" the larger string.  If shorter string is the
    // the second string, then return a positive value which equals the
    // first character in s1 after s2 ends.  If shorter string is the first
    // string, return a negative value which equals the first character in
    // s2 after s1 ends. 
    if (len2 < len1)
      return (int)s1[len2];   
    else if (len1 < len2)
      return -((int)s2[len1]);   

    // otherwise strings match 
    return 0;
}
#endif



static double double_tok(char *first_tok);
static int    int_tok(char* first_tok);
static bool   bool_tok(char *first_tok);


class MuseParameters  {
  public:	
    MuseParameters();	
    virtual ~MuseParameters(){};	

    bool read_parameters(char *parameter_file);

    int win_rows() { return rows; };
    int win_cols() { return cols; };
    int p() { return p_val; };
    int min_feature() { return Min_feature; };
    int extract_number() { return Extract_Num; };

    double ignore_pct() { return ignore_percentage; };
    double mu() { return mu_val; };
    double mu_int() { return mu_int_val; };

    bool is_interior() { return interior_mode; };

    // calculate the number of random samples needed
    int samples(int num_fits, int pts, int pts_remain);
    
  protected:	
    int rows, cols,      // size of the window
	p_val,           // number of point to instantiate a fit
	Min_feature,     // minimum feature size
	Extract_Num;     // number of features to extract
    double ignore_percentage;   // percentage of the residuals to ignore,
                         // this is converted to integer k0 for each surface
    double mu_val,       // multiplier on sigma_hat to find inliers
	mu_int_val;      // multiplier on sigma_hat to find inliers to
                         // an interior fit.
    double data_missing; // value of a bad range pixel
    bool interior_mode;

    // variables needed for determining the number of random samples
    double Smp_Max_Out_Pcnt;        // estimated maximum outlier percentage
    int Smp_Max_Surfaces,           // estimate maximum number of surfaces
        Smp_Min_Samples;            // minimum number of random samples to draw
    double Prob_Good_Pts_Found;     // prob of select a good point
    
};


MuseParameters::MuseParameters()
          : rows( 11 ),
            cols( 11 ),
	    p_val ( 3 ),            // planes
	    ignore_percentage( 0.10 ),
	    mu_val( 2.0 ),
	    mu_int_val( 3.0 ),
	    interior_mode( false ),
	    data_missing( -320.0 ),
	    Min_feature( 10 ),
	    Extract_Num( 2 ),
	    Smp_Max_Out_Pcnt( 0.10 ),
	    Smp_Max_Surfaces( 10 ),
	    Smp_Min_Samples( 10 ),
	    Prob_Good_Pts_Found( 0.99 )
{
}


//
// This is pretty much pulled out of Chuck Stewart's MINPRAN code, I think
// the only changes are the lines setting Smp_Min_N to Min_feature, and P
// to p_val. These changes were just to adapt my variables to Chuck's code.
//
int MuseParameters::samples(int num_fits, int pts, int pts_remain )
{
  float q;  /** the probability that a sample is good **/
  float ratio;
  int  min_pts, surf_remain, samples, i, bad;

  int Smp_Min_N = Min_feature;
  int P = p_val;
  
  if (pts_remain < Smp_Min_N) return(0);
  bad = (int) floor(pts*Smp_Max_Out_Pcnt);
  surf_remain =
    MY_MIN( Smp_Max_Surfaces-num_fits,
           (int) floor( (double) (pts_remain - bad) / Smp_Min_N));
  
  if (surf_remain < 1) {
    if ( pts_remain == Smp_Min_N ) return( Smp_Min_Samples );
    surf_remain=1;
    min_pts = MY_MAX(Smp_Min_N, pts_remain-bad);
  }
  else {
    min_pts = (int) floor( (double) (pts_remain - bad) / surf_remain );
  }
  
  ratio = 1.0;
  for ( i=0; i<P; i++ ) {
    ratio = ratio * ((float) min_pts - i) / ((float) pts_remain - i);
  }
  q = surf_remain * ratio;
  
  samples = (int)ceil( log10((double)(1.0 - Prob_Good_Pts_Found)) /
                      log10((double)(1.0 - q)) );
  
  if (samples < Smp_Min_Samples) samples = Smp_Min_Samples;
  return(samples);
}



bool MuseParameters::read_parameters(char *parameter_file)
{
    ifstream infile(parameter_file, ios::in);
    const int LINE_LENGTH=256;
    char line[LINE_LENGTH], *first_tok;

    if (!infile)
    {
      cerr << "MuseParameters: cannot open parameter file "
	   << parameter_file << endl << "Exiting. " << endl;
	exit(0);
    }

    while ( infile.peek() != EOF )
    {
	// grab a line from the parameter file
	infile.getline(line, LINE_LENGTH);

	// check for comment character
	if (line[0] == '#')
	    continue;

	// otherwise try to parse
	first_tok = strtok(line, " \t");
	if (first_tok != NULL)
	{
	    // determine the token type and obtain the parameters
	    if (strcasecmp(first_tok, "MU") == 0)
	        mu_val = double_tok(first_tok);
	    else if (strcasecmp(first_tok, "MU_INT") == 0)
		mu_int_val = double_tok(first_tok);
	    else if (strcasecmp(first_tok, "IGNORE_PCT") == 0)
		ignore_percentage = double_tok(first_tok);
	    else if (strcasecmp(first_tok, "WINDOW_SIZE") == 0)
	    {
		rows = int_tok(first_tok);
		cols = int_tok(first_tok);
	    }
	    else if (strcasecmp(first_tok, "DATA_MISSING") == 0)
		data_missing = double_tok(first_tok);
	    else if (strcasecmp(first_tok, "INTERIOR") == 0)
		interior_mode = bool_tok(first_tok);
	    else if (strcasecmp(first_tok, "MIN_FEATURE") == 0)
		Min_feature = int_tok(first_tok);
	    else if (strcasecmp(first_tok, "EXTRACT_NUMBER") == 0)
		Extract_Num = int_tok(first_tok);
	    else if (strcasecmp(first_tok, "MAX_OUTLIER_PCT") == 0)
		Smp_Max_Out_Pcnt = double_tok(first_tok);
    	    else if (strcasecmp(first_tok, "WINDOW_MAX_SURFACES") == 0)
		Smp_Max_Surfaces = int_tok(first_tok);
	    else if (strcasecmp(first_tok, "MIN_SAMPLES") == 0)
		Smp_Min_Samples = int_tok(first_tok);
	    else if (strcasecmp(first_tok, "PROB_GOOD_SAMPLE") == 0)
		Prob_Good_Pts_Found = double_tok(first_tok);
	}
    }

    return true;
}



//
// grab the next token and convert it a double,
//     first_tok is used for error reporting, presumably first_tok
//     is a token indicating the "name" of a parameter and double_tok()
//     is being called to extract the next token which would be the
//     double that is to be assigned to the first_tok parameter
//
double double_tok(char *first_tok)
{
    char *second_tok;

    
    second_tok = strtok(NULL, " \t");
    if (second_tok == NULL)
    {
	cerr << "Parameter file: no value found after token "
	     << first_tok << endl;
	return HUGE_VAL;
    }

    return atof(second_tok);
}


//
// grab the next token and convert it a int,
//     first_tok is used for error reporting (see double_tok)
//
int int_tok(char *first_tok)
{
    char *second_tok;

    
    second_tok = strtok(NULL, " \t");
    if (second_tok == NULL)
    {
	cerr << "Parameter file: no value found after token "
	     << first_tok << endl;
	return MAXINT;
    }

    return atoi(second_tok);
}


//
// grab the next token and convert it a bool,
//     first_tok is used for error reporting (see double_tok)
//
bool bool_tok(char *first_tok)
{
    char *second_tok;

    
    second_tok = strtok(NULL, " \t");
    if (second_tok == NULL)
    {
	cerr << "Parameter file: no value found after token "
	     << first_tok << endl;
	return false;
    }

    if (strcasecmp(second_tok, "true") == 0)
	return true;
    else if (strcasecmp(second_tok, "false") == 0)
	return false;
    
    return false;
}





#endif /* _MUSEPARAMETERS_H */


