
#ifndef _PIXELSTAT_H
#define _PIXELSTAT_H

/************************************************************************
 *
 * File: PixelStat.h
 *
 * Copyright (c) 1995 by Rensselaer Polytechnic Institute,
 *                       Troy, NY 12180
 *
 * Created:       Fri Oct  6 12:40:19 1995 by James Miller
 * Last Modified: Mon Dec 11 11:19:21 1995 by James Miller
 *
 * This class holds a variety of data pertaining to a range pixel.
 * For instance, it remembers the row and column of a pixel, a pointer
 * to the Vector3 for that pixel, and has placeholders for a residual
 * distance, a sigma estimate, and a surface pointer.
 *
 * There is also a field indicating whether the pixel is available for
 * operations (i.e. has it already been "visited/classified")
 *
 ************************************************************************/

#ident "@(#)PixelStat.h	1.4 95/12/11 RPI"

#include <stddef.h>

#ifndef bool
#include <bool.h>
#endif

class Vector3;
class Surface;

class PixelStat {
  public:
    PixelStat();
    PixelStat(int row_pos, int col_pos, const Vector3* pos);
    PixelStat(const PixelStat &px);
    virtual ~PixelStat(){};	

    // read access routines
    int row() const { return rowval; };
    int col() const { return colval; };
    bool available() const { return avail; };
    double r() const { return rval; };
    double s() const { return sval; };
    const Vector3* position() const { return pt; };
    const Surface* surface() const { return surf; };

    // write access routines
    PixelStat& row(int row_pos) { rowval = row_pos; return *this; };
    PixelStat& col(int col_pos) { colval = col_pos; return *this; };
    PixelStat& available(bool a) { avail = a; return *this; };
    PixelStat& r(double rv) { rval = rv; return *this; };
    PixelStat& s(double sv) { sval = sv; return *this; };
    PixelStat& position(const Vector3* p) {pt = p; return *this; };
    PixelStat& surface(const Surface* s) {surf = s; return *this; };

    PixelStat& clear();

    // operator== only tests if the row and columns match
    bool operator==(const PixelStat &px) const
                 { return ((px.rowval == rowval) && (px.colval == colval)); };
    bool operator!=(const PixelStat &px) const
                 { return ((px.rowval != rowval) || (px.colval != colval)); };
    
  protected:	
    int rowval, colval;
    bool avail;
    double rval, sval;
    const Vector3 *pt;
    const Surface *surf;

};

inline PixelStat::PixelStat()
{
    rowval = 0; colval = 0;
    avail  = true;
    rval   = 0.0;
    sval   = 0.0;
    pt     = NULL;
    surf   = NULL;    
}

inline PixelStat::PixelStat(int row_pos, int col_pos, const Vector3* pos)
{
    rowval = row_pos;   colval = col_pos;
    avail  = true;
    rval   = 0.0;
    sval   = 0.0;
    pt     = pos;
    surf   = NULL;
}

inline PixelStat::PixelStat(const PixelStat &px)
{
    rowval = px.rowval;   colval = px.colval;
    avail  = px.avail;
    rval   = px.rval;
    sval   = px.sval;
    pt     = px.pt;
    surf   = px.surf;
}


// empty/reset the values stored
inline PixelStat& PixelStat::clear()
{
    rowval = 0; colval = 0;
    avail  = true;
    rval   = 0.0;
    sval   = 0.0;
    pt     = NULL;
    surf   = NULL;        
    
    return *this;
}


#endif /* _PIXELSTAT_H */







