
/************************************************************************
 *
 * File: Plane.cc
 *
 * Copyright (c) 1995 by Rensselaer Polytechnic Institute,
 *                       Troy, NY 12180
 *
 * Created:       Tue Oct 10 13:39:52 1995 by James Miller
 * Last Modified: Mon Apr  8 11:29:26 1996 by James Miller
 *
 * Surface class for defining a plane.
 *            N*P + d = 0;
 *
 ************************************************************************/

#ident "@(#)Plane.cc	1.6 96/04/08 RPI"

#include <vector.h>
#include <algo.h>
#include "Vector3.h"
#include "Plane.h"
#include "misc.h"


// Tolerance for singular values in double precision
#define TOL 1.0e-10


Plane::Plane() : Surface(),
                 N(0, 0, 1),  // default to xy plane
                 d(0.0)
{
}


Plane::Plane(const Vector3 &Normal, double d_intersept)
           : Surface(),
	     N( Normal )
{
    double len;
    
    d = d_intersept;

    // check to make sure N is a unit vector
    len = N.norm();
    if (len != 1.0)
	N /= len;
}


Plane::Plane(const Vector3 &Normal, const Vector3 &P0)
          : Surface(),
	    N( Normal )
{
    double len;
    
    // check to make sure N is a unit vector
    len = N.norm();
    if (len != 1.0)
	N /= len;

    // N*(P-P0) = 0, do d = -N*P0
    d = -(N*P0);

}

Plane::Plane(const Vector3 &pt1, const Vector3 &pt2, const Vector3 &pt3)
             : Surface()
{
    double len;
    
    // pt1, pt2, pt3 are three points in the plane, normal is cross product
    // of two vectors on the plane
    N = ((pt2 - pt1)) % (pt3 - pt1);

    // check to make sure N is a unit vector
    len = N.norm();
    if (len != 1.0)
	N /= len;

    // N*(P-P0) = 0, do d = -N*P0
    d = -(N*pt1);
}


// call the other surface constructor to grab plane's distance metric setting
Plane::Plane(const Plane &plane)
           : Surface( plane ),
	     N ( plane.N )
{
    d = plane.d;
}


Plane::~Plane()
{
}


Plane& Plane::normal(const Vector3 &Normal)
{
    double len;
    
    N = Normal;

    // check to make sure N is a unit vector
    len = N.norm();
    if (len != 1.0)
	N /= len;

    return *this;
}


Vector3 Plane::normal() const
{
    Vector3 ret(N);

    return ret;
}



Plane& Plane::define(const Vector3 &pt1, const Vector3 &pt2,const Vector3 &pt3)
{
    double len;
    
    // pt1, pt2, pt3 are three points in the plane, normal is cross product
    // of two vectors on the plane
    N = ((pt2 - pt1)) % (pt3 - pt1);

    // check to make sure N is a unit vector
    len = N.norm();
    if (len != 1.0)
	N /= len;

    // N*(P-P0) = 0, do d = -N*P0
    d = -(N*pt1);

    return *this;
}



//
// Least squares fitting routine, return value is the standard deviation
//
double Plane::ls_fit(vector<const Vector3*> &pts )
{
    int i;
    double sigma = HUGE_VAL;
    double *W, *b, *a;
    double **X, **V;
    double wmax, thresh;

    double dist;
    vector<const Vector3*>::iterator pt_it;
    const Vector3 *pt_ptr;
    Vector3 v;
    
    switch ( metric )
    {
      case ordinary:
	// Perform the least squares fit using Singular Value Decomposition

	// allocate the matrices
	X = dmatrix(1, pts.size(), 1, 3);
	W = dvector(1, 3);
	V = dmatrix(1, 3, 1, 3);
	b = dvector(1, pts.size());
	a = dvector(1, 3);
	
	// fill X and b
	for (pt_it = pts.begin(), i=1; pt_it != pts.end(); ++pt_it, ++i)
	{
	    pt_ptr = *pt_it;
	    X[i][1] = 1.0;  X[i][2] = pt_ptr->x(); X[i][3] = pt_ptr->y();
	    b[i] = pt_ptr->z();
	}

	// compute the Singular Value Decomposition,
	//     note X gets overwritten by the values of U
	svdcmp(X, pts.size(), 3, W, V);

	// Edit the near singular values, values less than TOL*wmax,
	//    this provides stability in the parameter values
	wmax = 0.0;
	for (i=1; i<=3; i++)
	    if (W[i] > wmax) wmax = W[i];
	thresh = TOL*wmax;
	for (i=1; i<=3; i++)
	    if (W[i] < thresh) W[i] = 0.0;

	// compute the parameter values via backsubstitution
	svbksb(X, W, V, pts.size(), 3, b, a);

	// the vector "a" now holds the fit parameters, which we need to
	// convert (and store) as N*P + d = 0
	v.x( a[2] ).y( a[3] ).z( -1.0 );
	intersept( a[1] / v.norm() );
	normal( v );                // storing the normal normalizes the vector

	// all that is left is to compute the sigma estimate of the data
	// This method simply computes the standard deviation of the residuals
	// to the fit as sqrt( sum( r^2 )/n ).
	//
	// Here, we will use the current distance metric, so this code
	// can be used for the other fitting procedures as well.
	sigma = 0;
	for (pt_it = pts.begin(); pt_it != pts.end(); ++pt_it)
	{
	    dist = distance( *(*pt_it) );
	    sigma += (dist*dist);
	}
	sigma = sqrt( sigma / (double)pts.size());
	
	// deallocate the matrices
	free_dmatrix(X, 1, pts.size(), 1, 3);
	free_dmatrix(V, 1, 3, 1, 3);
	free_dvector(W, 1, 3);
	free_dvector(b, 1, pts.size());
	free_dvector(a, 1, 3);
	
	
	break;

      default:
	cerr << "Plane: least squares fits are currently only defined "
	     << "for ordinary regression." << endl;
	exit(0);
    }
    
    return sigma;
}


//
// Least squares fitting routine, return value is the standard deviation.
// This version overwrites the second parameter with the matrix (X'X)^(-1)
// and overwrites the third parameter with the vector X1mean (where
// X(nxq) = [1 X1]). XtXinv and X1mean are "Numerical Recipes in C matrices"
// thus they are indexed from 1.
//
// This version of ls_fit() has been altered so that matrices are only
// allocated if sizes increase.  Care must be taken when adding other
// fitting methods, since we do not check for changes in W, V, or a.
// For that matter, we currenly only change the sizes based on n.
// (Note, we overallocate.)
//
// Again, we have to rely on the memory manager to free this memory upon
// exiting.
//
//
double Plane::ls_fit(vector<const Vector3*> &pts,
		     double *const *const XtXinv, double *X1mean )
{
    int i, j, ii;
    double sigma = HUGE_VAL;
    static double *W, *b, *a = NULL;
    static double **X = NULL, **V;
    static double **VW;
    static int capacity = 0;
    double wmax, thresh;

    static double **XtX = NULL, *Xty = NULL, **XtX_temp = NULL, *col = NULL;
    static int *indx;
    double d;
    
    double dist;
    vector<const Vector3*>::iterator pt_it;
    const Vector3 *pt_ptr;
    Vector3 v;

    switch ( metric )
    {
      case ordinary:
	if (pts.size() < 1000)
	{
	  // Perform the least squares fit using Singular Value Decomposition

	  if (X == NULL) {
	    // allocate the matrices
	    X  = dmatrix(1, pts.size(), 1, 3);
	    W  = dvector(1, 3);
	    V  = dmatrix(1, 3, 1, 3);
	    VW = dmatrix(1, 3, 1, 3);
	    b  = dvector(1, pts.size());
	    a  = dvector(1, 3);
	    capacity = pts.size();
	  }
	  else if (pts.size() > capacity) {
	    // reallocate the matrices
	    X = re_dmatrix(X, 1, capacity, 1, 3, 1, 2*pts.size(), 1, 3);
	    b = re_dvector(b, 1, capacity, 1, 2*pts.size());
	    capacity = 2*pts.size();
	  }
	
	  // fill X,  b and X1mean
	  X1mean[1] = X1mean[2] = 0.0;
	  for (pt_it = pts.begin(), i=1; pt_it != pts.end(); ++pt_it, ++i)
	  {
	    pt_ptr = *pt_it;
	    X[i][1] = 1.0;  X[i][2] = pt_ptr->x(); X[i][3] = pt_ptr->y();
	    b[i] = pt_ptr->z();

	    X1mean[1] += X[i][2];   X1mean[2] += X[i][3];
	  }
	  X1mean[1] /= pts.size();   X1mean[2] /= pts.size();

	  // compute the Singular Value Decomposition,
	  //     note X gets overwritten by the values of U
	  svdcmp(X, pts.size(), 3, W, V);

	  // Edit the near singular values, values less than TOL*wmax,
	  //    this provides stability in the parameter values
	  wmax = 0.0;
	  for (i=1; i<=3; i++)
	    if (W[i] > wmax) wmax = W[i];
	  thresh = TOL*wmax;
	  for (i=1; i<=3; i++)
	    if (W[i] < thresh) W[i] = 0.0;

	  // compute the parameter values via backsubstitution
	  svbksb(X, W, V, pts.size(), 3, b, a);

	  // the vector "a" now holds the fit parameters, which we need to
	  // convert (and store) as N*P + d = 0
	  v.x( a[2] ).y( a[3] ).z( -1.0 );
	  intersept( a[1] / v.norm() );
	  normal( v );             // storing the normal normalizes the vector

	  // all that is left is to compute the sigma estimate of the data
	  // This method computes the standard deviation of the residuals
	  // to the fit as sqrt( sum( r^2 )/n ).
	  //
	  // Here, we will use the current distance metric, so this code
	  // can be used for the other fitting procedures as well.
	  sigma = 0.0;
	  for (pt_it = pts.begin(); pt_it != pts.end(); ++pt_it)
	  {
	    dist = distance( *(*pt_it) );
	    sigma += (dist*dist);
	  }
	  sigma = sqrt( sigma / (double)pts.size());

	  // in this version of ls_fit(), we need to calculate the matrix
	  // (X'X)^(-1), or the covariance matrix of the dependent variables.
	  // SVD returns the square roots of the eigenvalues of X'X and
	  // the eigenvectors of X'X.  Therefore the inverse of X'X can
	  // be reconstructed directly from these matrices.
	  //
	  // X = U W V', so X'X = V W' U' U W V' = V W'W V'
	  // and (X'X)^(-1) = V (W'W)^(-1) V'
	
	  // First, invert W'W, this squares and inverts the non-zero W's.
	  // W is overwritten
	  for (i=1; i<=3; i++)
	    if (W[i] != 0.0) W[i] = 1.0/(W[i]*W[i]);

	  // Now multiply V (W'W)^(-1) V'
	  for (i=1; i<=3; i++)
	    for (j=1; j<=3; j++) {
	      XtXinv[i][j] = 0.0;
	      VW[i][j] = V[i][j]*W[j];
	    }
	  for (i=1; i<=3; i++)
	    for (j=1; j<=3; j++)
	      for (ii=1; ii<=3; ii++)
		XtXinv[i][j] += VW[i][ii]*V[j][ii];
	
	
	  // deallocate the matrices
	  //free_dmatrix(X, 1, pts.size(), 1, 3);
	  //free_dmatrix(V, 1, 3, 1, 3);
	  //free_dmatrix(VW, 1, 3, 1, 3);
	  //free_dvector(W, 1, 3);
	  //free_dvector(b, 1, pts.size());
	  //free_dvector(a, 1, 3);
	}
	else
	{
	  // Perform the least squares fit using normal equations
	  //    faster for big surfaces

	  if (XtX == NULL) {
	    // allocate the static data
	    XtX = dmatrix(1, 3, 1, 3);
	    XtX_temp = dmatrix(1, 3, 1, 3);
	    Xty = dvector(1, 3);
	    indx = ivector(1, 3);
	    col = dvector(1, 3);
	  }
	  if (a == NULL) a = dvector(1, 3);

	  // clear the matrices
	  X1mean[1] = X1mean[2] = 0.0;
	  for (i=1; i<=3; i++) {
	    Xty[i] = 0.0;
	    a[i] = 0.0;
	    for (j=1; j<=3; j++) {
	      XtX[i][j] = 0.0;
	    }
	  }

	  // build XtX and Xty, note that we make use of the symmetry in XtX
	  XtX[1][1] = (double)pts.size();
	  for (pt_it = pts.begin(); pt_it != pts.end(); pt_it++)
	  {
	    pt_ptr = *pt_it;
	    Xty[1] += pt_ptr->z();
	    Xty[2] += pt_ptr->x() * pt_ptr->z();
	    Xty[3] += pt_ptr->y() * pt_ptr->z();

	    X1mean[1] += pt_ptr->x();
	    X1mean[2] += pt_ptr->y();
	    
	    XtX[2][2] += pt_ptr->x() * pt_ptr->x();
	    XtX[2][3] += pt_ptr->x() * pt_ptr->y();
	    XtX[3][3] += pt_ptr->y() * pt_ptr->y();
	  }
	  XtX[3][2] = XtX[2][3];

	  XtX[1][2] = X1mean[1]; XtX[1][3] = X1mean[2];
	  XtX[2][1] = X1mean[1]; XtX[3][1] = X1mean[2];
	  
	  X1mean[1] /= (double)pts.size();  X1mean[2] /= (double)pts.size();
	  
	  // calculate XtXinv
	  for (i=1; i<=3; i++)
	    for (j=1; j<=3; j++)
	      XtX_temp[i][j] = XtX[i][j];
	  ludcmp(XtX_temp, 3, indx, &d);    // overwrite XtX_temp
	  for (j=1; j<=3; j++) {
	    for (i=1; i<=3; i++) col[i] = 0.0;
	    col[j] = 1.0;
	    lubksb(XtX_temp, 3, indx, col);
	    for(i=1; i<=3; i++) XtXinv[i][j] = col[i];
	  }

	  // calculate the fit
	  for (i=1; i<=3; i++)
	    for (j=1; j<=3; j++)
	      a[i] += XtXinv[i][j]*Xty[j];
	  v.x( a[2] ).y( a[3] ).z( -1.0 );
	  intersept( a[1] / v.norm() );
	  normal( v );            // storing the normal normalizes the vector

	  // calculate sigma, for numerical stability, we calculate sigma
	  // via the actual residual distances
	  sigma = 0.0;
	  for (pt_it = pts.begin(); pt_it != pts.end(); ++pt_it)
	  {
	    dist = distance( *(*pt_it) );
	    sigma += (dist*dist);
	  }
	  sigma = sqrt( sigma / (double)pts.size());
  

	  // deallocate the matrices
	  //free_dmatrix(XtX, 1, 3, 1, 3);
	  //free_dmatrix(XtX_temp, 1, 3, 1, 3);
	  //free_dvector(Xty, 1, 3);
	  //free_dvector(col, 1, 3);
	  //free_ivector(indx, 1, 3);
	  //free_dvector(a, 1, 3);      // careful, this is same "a" as in svd
	}
	break;

      default:
	cerr << "Plane: least squares fits are currently only defined "
	     << "for ordinary regression." << endl;
	exit(0);
    }

    return sigma;
}




//
// Least squares fitting routine, return value is the standard deviation.
// This version overwrites the second parameter with the matrix (X'X)^(-1)
// and overwrites the third parameter with the vector X1mean (where
// X(nxq) = [1 X1]). XtXinv and X1mean are "Numerical Recipes in C matrices"
// thus they are indexed from 1.
//
// This version of ls_refit() has been altered so that matrices are only
// allocated if sizes increase.  Care must be taken when adding other
// fitting methods, since we do not check for changes in W, V, or a.
// For that matter, we currenly only change the sizes based on n.
// (Note, we overallocate.)
//
// Again, we have to rely on the memory manager to free this memory upon
// exiting.
//
// This version will also accelerate a refit if possible.  If a fit contains
// enough points and if this surface was used in the previous refit, then
// cached version of XtX and Xty will be used. This "accerlation" using
// cached results assumes that all new points are at the end of the
// pts lists AND that no points have been deleted from the list.
//
//
double Plane::ls_refit(vector<const Vector3*> &pts,
		       double *const *const XtXinv, double *X1mean )
{
    int i, j;
    double sigma = HUGE_VAL;

    static double **XtX = NULL, *Xty = NULL, **XtX_temp = NULL, *col = NULL;
    static double *a = NULL;
    static int *indx;
    double d;
    
    double dist;
    vector<const Vector3*>::iterator pt_it;
    const Vector3 *pt_ptr;
    Vector3 v;

    static Plane *previous_surface = NULL;
    static int old_n = 0;
    
    switch ( metric )
    {
      case ordinary:
	if (pts.size() < 1000)
	{
	  // Perform the least squares fit using Singular Value Decomposition

	  // no caching of result provided
	  previous_surface = NULL;
	  old_n = 0;
	  return ls_fit(pts, XtXinv, X1mean);
	}
	else
	{
	  // Perform the least squares fit using normal equations
	  //    use cached results if possible

	  if (XtX == NULL) {
	    // allocate the static data
	    XtX = dmatrix(1, 3, 1, 3);
	    XtX_temp = dmatrix(1, 3, 1, 3);
	    Xty = dvector(1, 3);
	    indx = ivector(1, 3);
	    col = dvector(1, 3);
	    a = dvector(1, 3);
	  }

	  if (previous_surface != this)
	  {
	    // cannot used cached results.

	    // clear the matrices
	    X1mean[1] = X1mean[2] = 0.0;
	    for (i=1; i<=3; i++) {
	      Xty[i] = 0.0;
	      for (j=1; j<=3; j++) {
		XtX[i][j] = 0.0;
	      }
	    }
	    old_n = 0;
	  }
	  else
	  {
	    // use cached results
	    assert( pts.size() >= old_n );
	    X1mean[1] *= (double)old_n;    X1mean[2] *= (double)old_n;

	    cout << "\tUsing cached results. ";
	  }
	  
	  // build XtX and Xty, note that we make use of the symmetry in XtX
	  XtX[1][1] = (double)pts.size();
	  for (pt_it = pts.begin()+old_n; pt_it != pts.end(); pt_it++)
	  {
	    pt_ptr = *pt_it;
	    Xty[1] += pt_ptr->z();
	    Xty[2] += pt_ptr->x() * pt_ptr->z();
	    Xty[3] += pt_ptr->y() * pt_ptr->z();

	    X1mean[1] += pt_ptr->x();
	    X1mean[2] += pt_ptr->y();
	    
	    XtX[2][2] += pt_ptr->x() * pt_ptr->x();
	    XtX[2][3] += pt_ptr->x() * pt_ptr->y();
	    XtX[3][3] += pt_ptr->y() * pt_ptr->y();
	  }
	  XtX[3][2] = XtX[2][3];

	  XtX[1][2] = X1mean[1]; XtX[1][3] = X1mean[2];
	  XtX[2][1] = X1mean[1]; XtX[3][1] = X1mean[2];
	  
	  X1mean[1] /= (double)pts.size();  X1mean[2] /= (double)pts.size();

	  // calculate XtXinv
	  for (i=1; i<=3; i++)
	    for (j=1; j<=3; j++)
	      XtX_temp[i][j] = XtX[i][j];
	  ludcmp(XtX_temp, 3, indx, &d);    // overwrite XtX_temp
	  for (j=1; j<=3; j++) {
	    for (i=1; i<=3; i++) col[i] = 0.0;
	    col[j] = 1.0;
	    lubksb(XtX_temp, 3, indx, col);
	    for(i=1; i<=3; i++) XtXinv[i][j]  = col[i];
	  }

	  // calculate the fit
	  for (i=1; i<=3; i++) {
	    a[i] = 0.0;
	    for (j=1; j<=3; j++)
	      a[i] += XtXinv[i][j]*Xty[j];
	  }
	  v.x( a[2] ).y( a[3] ).z( -1.0 );
	  intersept( a[1] / v.norm() );
	  normal( v );            // storing the normal normalizes the vector

	  // calculate sigma, for numerical stability, we calculate sigma
	  // via the actual residual distances
	  sigma = 0.0;
	  for (pt_it = pts.begin(); pt_it != pts.end(); ++pt_it)
	  {
	    dist = distance( *(*pt_it) );
	    sigma += (dist*dist);
	  }
	  sigma = sqrt( sigma / (double)pts.size());
  
	  // deallocate the matrices
	  //free_dmatrix(XtX, 1, 3, 1, 3);
	  //free_dmatrix(XtX_temp, 1, 3, 1, 3);
	  //free_dvector(Xty, 1, 3);
	  //free_dvector(col, 1, 3);
	  //free_ivector(indx, 1, 3);
	  //free_dvector(a, 1, 3);     
	}
	break;

      default:
	cerr << "Plane: least squares fits are currently only defined "
	     << "for ordinary regression." << endl;
	exit(0);
    }

    previous_surface = this;
    old_n = pts.size();
    return sigma;
}


#undef TOL
