
#ifndef _PLANE_H
#define _PLANE_H

/************************************************************************
 *
 * File: Plane.h
 *
 * Copyright (c) 1995 by Rensselaer Polytechnic Institute,
 *                       Troy, NY 12180
 *
 * Created:       Fri Oct  6 13:08:54 1995 by James Miller
 * Last Modified: Mon Apr  8 11:24:58 1996 by James Miller
 *
 * Surface class for defining a plane.
 *             N*P + d = 0;
 *
 ************************************************************************/

#ident "@(#)Plane.h	1.4 96/04/08 RPI"

#include <math.h>
#include "Surface.h"

class Vector3;
template <class T> class vector;

class Plane : public Surface {
  public:	
    Plane();
    Plane(const Vector3 &Normal, double d_intersept);
    Plane(const Vector3 &Normal, const Vector3 &P0);
    Plane(const Vector3 &pt1, const Vector3 &pt2, const Vector3 &pt3);
    Plane(const Plane &plane);
    
    virtual ~Plane();

    // set/get routines
    Plane& normal(const Vector3 &Normal);
    Vector3 normal() const;

    Plane& intersept( double d_intersept );
    double intersept() const;

    Plane& define(const Vector3 &pt1, const Vector3 &pt2, const Vector3 &pt3);
        
    // distance metrics
    double ordinary_distance(const Vector3 &pt) const;

    // projection routines
    Vector3 ordinary_project(const Vector3 &pt) const;
    
    // perform least-squares fit using the current distance metric
    // return value is standard deviation
    virtual double ls_fit(vector<const Vector3*>& pts);
    virtual double ls_refit(vector<const Vector3*> &pts) {return ls_fit(pts);};
  
    // perform least-squares fit using the current distance metric
    // return value is standard deviation, second parameter is overwritten
    // with the matrix (X'X)^(-1), third parameter is overwritten with the
    // vector X1mean (where X(nxq) = [1 X1]).  The matrix XtXinv and vector
    // X1mean must be preallocated for the proper size.  Note XtXinv and
    // X1mean are "Numerical Recipes in C matrices" so they are indexed
    // started at 1.
    virtual double ls_fit(vector<const Vector3*>& pts,
			  double *const *const XtXinv,
			  double *X1mean);
    virtual double ls_refit(vector<const Vector3*>& pts,
			    double *const *const XtXinv,
			    double *X1mean);
  protected:	
    Vector3 N;
    double  d;
    
};


//
// ordinary regression distance (i.e. error only in z) of a pt from
// the plane.
//
// Solve N*P + d = 0, for z
//        z = (-axi - byi - d)/c
//
// d = zi - z
//   = zi + (axi + byi + d)/c
//   = (N*P + d)/c
//
inline double Plane::ordinary_distance(const Vector3 &pt) const
{
    if (N.z() != 0.0)
	return (N*pt + d)/N.z();
    else                              // plane perpendicular to xy plane
    {
	if (N*pt + d == 0.0)          // pt is on plane
	    return 0.0;
	else                          // (xi, yi) is not on plane at all
	    return HUGE_VAL;
    }
}


//
// projection of the point pt onto the plane using ordinary regression
// projection direction (see ordinary_distance() above)
//
// z coordinate of the point projected on the surface is sol'n of
//        axi + byi + cz + d = 0
// for z.
//      
//
inline Vector3 Plane::ordinary_project(const Vector3 &pt) const
{
    Vector3 p;
    double z;

    if (N.z() != 0.0)
    {
	p = pt;
	z = -(N.x()*pt.x() + N.y()*pt.y() + d) / N.z();
	p.z( z );                // move pt along z back to the surface
	return p; 
    }
    else                              // plane perpendicular to xy plane
    {
	if (N*pt + d == 0.0)          // pt is already on plane
	    return pt;
	else                          // (xi, yi) is not on plane at all
	{
	    p = pt;
	    p.z( HUGE_VAL );          // not on plane, set z to infinity

	    return p;
	}
    }
    
}



inline Plane& Plane::intersept(double d_intersept)
{
    d = d_intersept;

    return *this;
}


inline double Plane::intersept() const
{
    return d;
}


//
//
//  Having trouble with stl and the Plane class.  "vector"'s static
//  allocator gets instantiate twice.  To fix this, I am placing Plane.cc
//  in the header file Plane.h and not building a Plane.o.  This is less
//  than ideal, but until I figure another way....
//
//
#include "Plane.cc"

#endif /* _PLANE_H */
