
/************************************************************************
 *
 * File: RangeImage.cc
 *
 * Copyright (c) 1995 by Rensselaer Polytechnic Institute,
 *                       Troy, NY 12180
 *
 * Created:       Thu Oct  5 16:24:41 1995 by James Miller
 * Last Modified: Mon Apr  8 11:33:32 1996 by James Miller
 *
 * Class to hold, create, read, and write a range image
 *
 ************************************************************************/

#ident "@(#)RangeImage.cc	1.4 96/04/08 RPI"

#include <iostream.h>
#include <fstream.h>
#include "Vector3.h"
#include "RangeImage.h"


// Allocate a range image, this allocates all the memory for a range image,
// so it is immediately available for accessing (reading/writing).
//
// Note that if read() follows the creation, the memory may be resized.
RangeImage::RangeImage(int rows, int cols) :
    invalid(-320.0),
    num_rows(rows),
    num_cols(cols),
    type(xyz)
{
    // allocate the image of pointers and all the necessary Vector3's
    allocate();
}


// Copy constructor, performs a deep copy... the new range image had
// identical type, size, and values (x,y,z), but is constructed of
// different Vector3 objects
RangeImage::RangeImage(const RangeImage &ri)
{
    // copy over the "stats"
    type = ri.type;
    invalid = ri.invalid;
    num_rows = ri.num_rows;
    num_cols = ri.num_cols;

    // create the new structure
    allocate();

    // copy the Vector3 values in...
    for(int i=0; i<num_rows*num_cols; i++)
	*(range_values[i]) = *(ri.range_values[i]);
    
}

RangeImage::~RangeImage()
{
    // free up all allocated memory
    deallocate();
}


//
// operator= performs a deep copy... the Vector3 values (not the pointers)
// are copied over
//
RangeImage& RangeImage::operator=(const RangeImage &ri)
{
    // copy over the "stats"
    type = ri.type;
    invalid = ri.invalid;

    // check if sizes currently match, if not, reallocate
    if ((num_rows != ri.num_rows) || (num_cols != ri.num_cols))
    {
	deallocate();         // free up whatever we had
	num_rows = ri.num_rows;    // store the new size
	num_cols = ri.num_cols;
	allocate();           // allocate the necessary memory
    }

    // Perform a deep copy on the Vector3's
    for(int i=0; i<num_rows*num_cols; i++)
	*(range_values[i]) = *(ri.range_values[i]);

    return *this;
}

//
// resize a range image, if size is different from before, all data is
// destroyed
//
RangeImage& RangeImage::resize(int rows, int cols)
{
    // check sizes currently match, if not, reallocate
    if ((num_rows != rows) || (num_cols != cols))
    {
	deallocate();         // free up whatever we had
	num_rows = rows;    // store the new size
	num_cols = cols;
	allocate();           // allocate the necessary memory
    }

    return *this;
}

//
// Read a range image.
//
// The invalid marker must be set before this call.
//
bool RangeImage::read(const char * name)
{
    ifstream infile(name, ios::in);
    streampos spos;
    int inrows, incols;
    int i,j;

    if (!infile)
    {
	cerr << "RangeImage: cannot open file " << name << endl;
	return false;
    }

    cerr << "RangeImage: reading range file " << name << endl;
    
    // peek inside the file for its size, if the size is large enough
    // to hold 3*rows*cols floats (plus header), then we have an xyz file,
    // otherwise we have a just_z file.

    // move to the end of the file
    infile.seekg(0, ios::end);

    // where are we?
    spos = infile.tellg();

    // move back to the beginning
    infile.seekg(0, ios::beg);

    // read the number of rows and columns
    infile.read((char *)&inrows, sizeof(int));
    if (infile.gcount() != sizeof(int))
    {
	cerr << "RangeImage: unable to read number of rows in image." << endl;
	return false;
    }

    infile.read((char *)&incols, sizeof(int));
    if (infile.gcount() != sizeof(int))
    {
	cerr << "RangeImage: unable to read number of cols in image." << endl;
	return false;
    }

    // Check file size against rows*cols
    if (spos == inrows*incols*sizeof(float) + 2*sizeof(int))
	type = just_z;
    else if (spos == inrows*incols*sizeof(float)*3 + 2*sizeof(int))
	type = xyz;
    else
    {
	cerr << "RangeImage: not enough data in the range file" << endl;
	return false;
    }

    // We have a valid range image... allocate the memory if necessary
    if ((num_rows != inrows) || (num_cols != incols))
    {
	deallocate();         // free up whatever we had
	num_rows = inrows;    // store the new size
	num_cols = incols;
	allocate();           // allocate the necessary memory
    }


    // allocate some temporary storage
    float *x = new float[num_rows*num_cols],
	  *y = new float[num_rows*num_cols],
	  *z = new float[num_rows*num_cols];
    if ((x == 0) || (y == 0) || (z == 0))
    {
	cerr << "RangeImage: not enough memory for "
	     << "temporary storage." << endl;
	return false;
    }


    // Read in the data
    if (type == xyz)
    {
	infile.read((char *)x, num_rows*num_cols*sizeof(float));
	if (infile.gcount() != num_rows*num_cols*sizeof(float))
	{
	    cerr << "RangeImage: file size changed while reading." << endl;
	    return false;
	}
	infile.read((char *)y, num_rows*num_cols*sizeof(float));
	if (infile.gcount() != num_rows*num_cols*sizeof(float))
	{
	    cerr << "RangeImage: file size changed while reading." << endl;
	    return false;
	}
	infile.read((char *)z, num_rows*num_cols*sizeof(float));
	if (infile.gcount() != num_rows*num_cols*sizeof(float))
	{
	    cerr << "RangeImage: file size changed while reading." << endl;
	    return false;
	}
    }
    else if (type == just_z)
    {
	// for just_z range files, we will convert row and column
	// positions into x and y positions, so that every range
	// image will have valid x,y,z's

	infile.read((char *)z, num_rows*num_cols*sizeof(float));
	if (infile.gcount() != num_rows*num_cols*sizeof(float))
	{
	    cerr << "RangeImage: file size changed while reading." << endl;
	    return false;
	}

	// dummy up the x and y positions
	for (i=0; i<num_rows; i++)
	    for (j=0; j<num_cols; j++)
	      if (z[i*num_cols + j] != invalid) {
		y[i*num_cols + j] = num_rows - float(i);		
		x[i*num_cols + j] = float(j);
	      }
	      else {
		x[i*num_cols + j] = invalid;
		y[i*num_cols + j] = invalid;
	      }
    }


    // all the data has been read into temporary store, now
    // stuff it into the range image
    Vector3 *pt;
    for (i = 0; i < num_rows*num_cols; i++)
    {
	pt = range_values[i];
	pt->x( x[i] );     // note we are implicitly converting float to double
	pt->y( y[i] );
	pt->z( z[i] );
    }


    // free up temporary storage
    delete [] x;
    delete [] y;
    delete [] z;

    infile.close();
    
    return true;
}


//
// Write a range image
//
bool RangeImage::write(const char* name) const
{
    ofstream outfile(name, ios::out);
    Vector3 *pt;
    float temp;
    int i;
    
    if (!outfile)
    {
	cerr << "RangeImage: cannot open file " << name << endl;
	return false;
    }

    cerr << "RangeImage: writing range file " << name << endl;

    // write out the number of rows and columns
    outfile.write( (char *)&num_rows, sizeof(int) );
    outfile.write( (char *)&num_cols, sizeof(int) );

    // write out all the x, then all the y, then all the z values
    if ( type == xyz )
    {
	for (i=0; i<num_rows*num_cols; i++)
	{
	    pt = range_values[i];
	    temp = pt->x();              // note this converts double to float
	    outfile.write( (char *)&temp, sizeof(float) );
	}
	for (i=0; i<num_rows*num_cols; i++)
	{
	    pt = range_values[i];
	    temp = pt->y();              // note this converts double to float
	    outfile.write( (char *)&temp, sizeof(float) );
	}
    }
    // whether xyz or just_z, write out the z image
    for (i=0; i<num_rows*num_cols; i++)
    {
	pt = range_values[i];
	temp = pt->z();              // note this converts double to float
	outfile.write( (char *)&temp, sizeof(float) );
    }

    outfile.close();
    
    return true;
}


// is a given range pixel "valid", if any component is valid, the entire
// point is valid (i.e. all components must be invalid for pt to be invalid)
bool RangeImage::is_valid(int row, int col) const
{
    Vector3 *pt = range_values[row*num_cols + col];

    return (! ((pt->x() == invalid) && (pt->y() == invalid)
	                         && (pt->z() == invalid)) );
}

// mark a given range pixel "invalid"
void RangeImage::mark_invalid(int row, int col)
{
    Vector3 *pt = (*this)[row][col];

    pt->x(invalid);
    pt->y(invalid);
    pt->z(invalid);
}


//
// Protected routine that will allocate the pointer image and each of
// of the necessary Vector3's
//
void
RangeImage::allocate()
{
    // allocate the array of pointers to Vector3's
    range_values = new Vector3*[num_rows*num_cols];
    assert(range_values != 0);
    
    // allocate each of the Vector3's
    for (int i=0; i<num_rows*num_cols; i++)
    {
	range_values[i] = new Vector3;
	assert(range_values[i] != 0);
    }
}


//
// Protected routine that will deallocate the pointer image and each of
// the allocated Vector3's
//
void
RangeImage::deallocate()
{
    // delete each of the Vector3's allocated
    for (int i=0; i<num_rows*num_cols; i++)
	delete range_values[i];
    
    // free up all the memory allocated to hold the pointers
    delete [] range_values;
}
