
#ifndef _RANGEIMAGE_H
#define _RANGEIMAGE_H

/************************************************************************
 *
 * File: RangeImage.h
 *
 * Copyright (c) 1995 by Rensselaer Polytechnic Institute,
 *                       Troy, NY 12180
 *
 * Created:       Thu Oct  5 15:50:29 1995 by James Miller
 * Last Modified: Mon Jan 29 10:38:13 1996 by James Miller
 *
 * Class to handle the creating, reading, writing, and manipulation
 * of a range image.
 *
 * Our range images are currently saved to disk in non-interleaved format,
 * i.e. after a header indicating the number of rows and columns, the
 * range data is organized as an image of all the x values, followed
 * by an image of all the y values, followed by an image of all the
 * z values.
 *
 * Should other formats come into play, there are two options:
 *      1.  Subclass off this object to read/write a new format
 *      2.  Add member functions to this class such the other
 *              formats can be handled.  For instance, the type of
 *              file could be specified either through direct
 *              member function calls or by passing a range type enum
 *              to the read and write methods.  This would also allow
 *              translating range image types.  If we get lucky, the
 *              read routine might be able to deduce the type of file
 *              directly without user interaction.
 *
 *
 * The range image is stored as a one-dimensional array of pointers to
 * Vector3 objects.  While it would have been possible to implement it
 * as strictly a one-dimensional array of Vector3's, the image of pointers
 * technique was chosen to ensure a little consistency between this and
 * other classes I will be using of similar form. This is subject to change.
 *
 * The operator[] is written to return a Vector3** which is the position
 * in the array corresponding to the beginning of a particular row.
 * Thus using a_range_image[row][col] will actually return the Vector3*
 * located at [ro][col] (the second subscript is handled implictly using
 * C array indexing.  I am not happy with this implementation since it
 * exposes the internal data. Thus the actual Vector3 can be changed
 * outside of the RangeImage class, something I would have preferred to
 * avoid.
 *
 * A word of warning given the above: while it is possible to change
 * the Vector3 data values outside of the RangeImage class, it is
 * important that one nevers "deletes" or frees this memory.  The
 * RangeImage created this memory and it is its responsibility to delete it.
 *
 ************************************************************************/

#ident "@(#)RangeImage.h	1.4 96/01/29 RPI"

#ifndef bool
#include <bool.h>
#endif

#include <assert.h>


class Vector3;


class RangeImage  {
  public:
    // enum for type of file
    enum range_type {xyz, just_z};
    
    // constructor/destructor
    RangeImage(int rows = 480, int cols = 512);
    RangeImage(const RangeImage &ri);            // performs a deep copy
    virtual ~RangeImage();	

    // a_range_image[row][col] will return a pointer to a Vector3.
    // The second function is for accessing constant RangeImages....
    Vector3** operator[](int row);
    const Vector3 *const * operator[](int row) const;
 
    RangeImage& operator=(const RangeImage &ri);   // performs a deep copy

    // resize a range image, this destroys the data if size actually changes
    RangeImage & resize(int rows, int cols);
    
    // read a range file
    bool read(const char * name);

    // write a range file
    bool write(const char * name) const;

    // setting the invalid pixel marker, testing a point for validity
    void set_invalid_marker(double val);
    double get_invalid_marker() const;
    bool is_valid(int row, int col) const;
    void mark_invalid(int row, int col);

    // setting/getting range image type
    void set_type(RangeImage::range_type mode) { type = mode; };
    RangeImage::range_type  get_type() const { return type; };

    // getting the size of the range image
    int rows() const { return num_rows; };
    int cols() const { return num_cols; };
    
  protected:
    // routines to allocate/deallocate the pointer image and the Vector3's
    void allocate();
    void deallocate();

    
    Vector3** range_values;    // the 2D array of range values are stored as 1D
                              // array of pointers to Vector3's
    int num_rows, num_cols;

    float invalid;            // value that a bad pixel is assigned

    range_type type;          // type of range image
};


// array indexing operator.  See notes above.
inline Vector3** RangeImage::operator[](int row)
{
    assert((row >= 0) && (row < num_rows));

    return (range_values + row*num_cols);
}


// array indexing operator for constant RangeImages.  
inline const Vector3 *const * RangeImage::operator[](int row) const
{
    assert((row >= 0) && (row < num_rows));

    return ((const Vector3 *const *)(range_values + row*num_cols));
}






// set the "value" which indicates an invalid range pixel...
inline void RangeImage::set_invalid_marker(double val)
{
    invalid = val;
}

// get the "value" which indicates an invalid range pixel...
inline double RangeImage::get_invalid_marker() const
{
    return invalid;
}





#endif /* _RANGEIMAGE_H */
