
#ifndef _SURFACE_H
#define _SURFACE_H

/************************************************************************
 *
 * File: Surface.h
 *
 * Copyright (c) 1995 by Rensselaer Polytechnic Institute,
 *                       Troy, NY 12180
 *
 * Created:       Fri Oct  6 13:05:50 1995 by James Miller
 * Last Modified: Mon Apr  8 11:26:26 1996 by James Miller
 *
 * Abstract class for holding a surface description and evaluating
 * the distance of a point to a surface.
 *
 * Need to make a decision as to whether distances can be signed.
 * Signed distances would indicate which side of surface a point lies
 * on.  For now, we'll assume distances are signed.
 *
 * Available distance metrics:
 *
 *    algebraic:
 *        Let f(x,y,z) = 0 represent the surface description, then
 *             d = f(xi, yi, zi)
 *
 *    ordinary:   (ordinary or usual regression distance)
 *        Let z = f(x,y) represent the surface description, then
 *             d = zi - f(xi, yi)
 *
 *    orthogonal:
 *        The shortest distance from the point (xi, yi, zi) to the
 *        surface (this distance is along a vector orthogonal to the
 *        surface).
 *
 *    line_of_sight:
 *        Distance of the point (xi, yi, zi) to the surface along the
 *        line of sight direction.  Here, we might assume that (xi, yi, zi)
 *        is along a line of sight vector from (0,0,0), i.e. there is no
 *        error in the spherical coordinates (phi, theta)
 *
 *    approximate_geometric:
 *        A normalization of the algebraic distance by the gradient
 *        of the algebraic distance.  This creates an approximate
 *        geometric distance that is suitable when points are close
 *        to a slowly varying surface.
 *
 * Class also has routines to "project" a point onto the surface, using
 * the current distance metric (direction), i.e. the routines return the
 * point on the surface that is closest to the given point along the
 * distance metric direction (might not be defined with algebraic distances).
 *
 * Note that all subclasses must implement the ordinary_distance()
 * and ordinary_project() methods.
 *
 ************************************************************************/

#ident "@(#)Surface.h	1.7 96/04/08 RPI"

#include <iostream.h>

class Vector3;
template <class T> class vector;

class Surface  {
 public:
  enum distance_metric {algebraic, ordinary, orthogonal, line_of_sight,
			approximate_geometric};
    
  Surface() { metric = ordinary; };	
  Surface(const Surface &surf) { metric = surf.metric; };
  virtual ~Surface(){};	

  // distance metrics
  virtual double ordinary_distance(const Vector3 &pt) const = 0;
  virtual double algebraic_distance(const Vector3 &pt) const;
  virtual double orthogonal_distance(const Vector3 &pt) const;
  virtual double line_of_sight_distance(const Vector3 &pt) const;
  virtual double approximate_geometric_distance(const Vector3 &pt) const;

  virtual double distance(const Vector3 &pt) const; // use current mode
    
  virtual void set_distance_metric(Surface::distance_metric dm);
  virtual Surface::distance_metric  get_distance_metric() const;

  // project a point onto the surface, i.e. return the point
  // on the surface that the given point projects to given the
  // current distance metric
  virtual Vector3 ordinary_project(const Vector3 &pt) const = 0;
  virtual Vector3 algebraic_project(const Vector3 &pt) const;
  virtual Vector3 orthogonal_project(const Vector3 &pt) const;
  virtual Vector3 line_of_sight_project(const Vector3 &pt) const;
  virtual Vector3 approximate_geometric_project(const Vector3 &pt) const;

  virtual Vector3 project(const Vector3 &pt) const; // use current mode

  // perform least-squares fit using the current distance metric
  // return value is standard deviation 
  virtual double ls_fit(vector<const Vector3*> &) = 0;
    
  // perform least-squares fit using the current distance metric
  // return value is standard deviation, second parameter is overwritten
  // with the matrix (X'X)^(-1), third parameter is overwritten with the
  // vector X1mean (where X(nxq) = [1 X1]).  The matrix XtXinv and vector
  // X1mean must be preallocated for the proper size.  Note XtXinv and
  // X1mean are "Numerical Recipes in C matrices" so they are indexed
  // started at 1.
  virtual double ls_fit(vector<const Vector3*>& pts,
			double *const *const XtXinv,
			double *X1mean) = 0;

  // make use of previously stored information if possible to accelerate
  // the refitting process.
  virtual double ls_refit(vector<const Vector3*> &);
  virtual double ls_refit(vector<const Vector3*>& pts,
			  double *const *const XtXinv,
			  double *X1mean);

    
 protected:	
  distance_metric metric;
};


inline void Surface::set_distance_metric(Surface::distance_metric dm)
{
    metric = dm;
}

inline Surface::distance_metric Surface::get_distance_metric() const
{
    return metric;
}


//
// Implementations for distance metrics occur in actual subclasses...
//

inline double Surface::algebraic_distance(const Vector3& pt) const
{
    cerr << "Surface: algebraic distance not implemented.  "
	 << "Defaulting to ordinary regression distances." << endl;

    return ordinary_distance(pt);
}


inline double Surface::orthogonal_distance(const Vector3& pt) const
{
    cerr << "Surface: orthogonal distance not implemented.  "
	 << "Defaulting to ordinary regression distances." << endl;

    return ordinary_distance(pt);
}


inline double Surface::line_of_sight_distance(const Vector3& pt) const
{
    cerr << "Surface: line of sight distance not implemented.  "
	 << "Defaulting to ordinary regression distances." << endl;

    return ordinary_distance(pt);
}


inline double Surface::approximate_geometric_distance(const Vector3& pt) const
{
    cerr << "Surface: approximate geometric distance not implemented.  "
	 << "Defaulting to ordinary regression distances." << endl;

    return ordinary_distance(pt);
}




//
// If this routine cannot be effectively inlined, I suggest using
// set_distance_metric() to (internally) store a pointer to the member 
// function for the current distance mode.  Then change this routine to
// an inline function where the switch statement is replaced with a
// call through the stored member function pointer. Unfortunately,
// this method prohibits distance functions to be virtual functions from
// multiply inherited objects.
//
// Keeping the switch statement here, however, will allow for the
// proper execution of virtual functions through multiple inheritiance
// (if necessary). But this method is potentially slower...
//
inline double Surface::distance(const Vector3& pt) const
{
    double d = HUGE_VAL;
    
    switch( metric )
    {
    case algebraic:     d = algebraic_distance(pt);      break;
    case ordinary:      d = ordinary_distance(pt);       break;
    case orthogonal:    d = orthogonal_distance(pt);     break;
    case line_of_sight: d = line_of_sight_distance(pt);  break;
    case approximate_geometric: d = approximate_geometric_distance(pt);
      break;
    }

    return d;
}


//
// Projection routines
//
// Again, actual implementations of projections occur in subclasses
//


inline Vector3 Surface::algebraic_project(const Vector3& pt) const
{
    cerr << "Surface: algebraic projection not implemented.  "
	 << "Defaulting to ordinary regression projection." << endl;

    return ordinary_project(pt);
}


inline Vector3 Surface::orthogonal_project(const Vector3& pt) const
{
    cerr << "Surface: orthogonal projection not implemented.  "
	 << "Defaulting to ordinary regression projection." << endl;

    return ordinary_project(pt);
}


inline Vector3 Surface::line_of_sight_project(const Vector3& pt) const
{
    cerr << "Surface: line of sight projection not implemented.  "
	 << "Defaulting to ordinary regression projection." << endl;

    return ordinary_project(pt);
}


inline Vector3 Surface::approximate_geometric_project(const Vector3& pt) const
{
    cerr << "Surface: approximate geometric projection not implemented.  "
	 << "Defaulting to ordinary regression projection." << endl;

    return ordinary_project(pt);
}




//
// The notes for this routine are similar to those for distance()
// 
inline Vector3 Surface::project(const Vector3& pt) const
{
    Vector3 p;
    
    switch( metric )
    {
    case algebraic:     p = algebraic_project(pt);      break;
    case ordinary:      p = ordinary_project(pt);       break;
    case orthogonal:    p = orthogonal_project(pt);     break;
    case line_of_sight: p = line_of_sight_project(pt);  break;
    case approximate_geometric: p = approximate_geometric_project(pt);
      break;
    }

    return p;
}


//
// Default refitting routines are to use the standard refit.  Override
// these if it is possible to cache results from a previous fit.
//
inline double Surface::ls_refit(vector<const Vector3*> &pts)
{
    return ls_fit(pts);
}

inline double Surface::ls_refit(vector<const Vector3*>& pts,
				double *const *const XtXinv,
				double *X1mean)
{
    return ls_fit(pts, XtXinv, X1mean);
}





#endif /* _SURFACE_H */
