
#ifndef _SURFACESTAT_H
#define _SURFACESTAT_H

/************************************************************************
 *
 * File: SurfaceStat.h
 *
 * Copyright (c) 1995 by Rensselaer Polytechnic Institute,
 *                       Troy, NY 12180
 *
 * Created:       Fri Oct 13 17:07:27 1995 by James Miller
 * Last Modified: Tue Mar 26 13:29:40 1996 by James Miller
 *
 * Class to hold data concerning a surface: a pointer to a surface,
 * its current sigma, pointers to vectors of points and pixels, and pointer
 * to a bounding box.
 *
 * SurfaceStat is a "collector" of information.  It is not responsible
 * for allocating or deallocating the surfaces or point lists that
 * it references. These objects (Surface, vector<Vector3*>,vector<PixelStat*>),
 * must be allocated and deallocated outside of this class.
 *
 ************************************************************************/

#ident "@(#)SurfaceStat.h	1.5 96/04/04 RPI"

#include <vector.h>

class Vector3;
class Surface;
class SurfaceStat;
class PixelStat;

class SurfaceStat {
 public:	
  SurfaceStat();
  SurfaceStat(Surface *surface, double s);
  SurfaceStat(Surface *surface, double s, vector<const Vector3*> *pt,
	      vector<PixelStat*> *px, Vector3 *b_box);
  SurfaceStat(const SurfaceStat &sf);
  virtual ~SurfaceStat(){};	

  // read access routines
  double sigma() const { return sigmaval; };
  Surface* surface()  { return surf; };
  vector<const Vector3*>* points() { return pts; };
  vector<PixelStat*>* pixels() {return pxs; };
  Vector3* boundingbox() {return bounding_box; };

  // write access routines
  SurfaceStat& sigma(double s) { sigmaval = s; return *this; };
  SurfaceStat& surface(Surface *s) { surf = s; return *this; };
  SurfaceStat& points(vector<const Vector3*> *p) { pts = p; return *this; };
  SurfaceStat& pixels(vector<PixelStat*> *p) {pxs = p; return *this;};
  SurfaceStat& boundingbox(Vector3 *b_box){bounding_box = b_box;return *this;};
  
  SurfaceStat& clear();

 protected:	
  Surface *surf;
  double sigmaval;
  vector<const Vector3*> *pts;    // list needed for final LS fit
  vector<PixelStat*> *pxs;        // list needed for projecting inliers
                                  // onto the surface
  Vector3 *bounding_box;
};


inline SurfaceStat::SurfaceStat()
{
    sigmaval = 0.0;
    surf = NULL;
    pts = NULL;
    pxs = NULL;
    bounding_box = NULL;
}


inline SurfaceStat::SurfaceStat(Surface *surface, double s)
{
    surf = surface;
    sigmaval = s;
    pts = NULL;
    pxs = NULL;
    bounding_box = NULL;
}

inline SurfaceStat::SurfaceStat(Surface *surface, double s,
				vector<const Vector3*> *pt,
				vector<PixelStat*> *px,
				Vector3 *b_box)
{
    surf = surface;
    sigmaval = s;
    pts = pt;
    pxs = px;
    bounding_box = b_box;
}

inline SurfaceStat::SurfaceStat(const SurfaceStat &sf)
{
    surf = sf.surf;
    sigmaval = sf.sigmaval;
    pts = sf.pts;
    pxs = sf.pxs;
    bounding_box = sf.bounding_box;
}


// empty/reset the values stored
inline SurfaceStat& SurfaceStat::clear()
{
    sigmaval = 0.0;
    surf = NULL;
    pts = NULL;
    pxs = NULL;
    bounding_box = NULL;
    
    return *this;
}

#endif /* _SURFACESTAT_H */


