
#ifndef _VECTOR3_H
#define _VECTOR3_H

/************************************************************************
 *
 * File: Vector3.h
 *
 * Copyright (c) 1995 by Rensselaer Polytechnic Institute,
 *                       Troy, NY 12180
 *
 * Created:       Thu Oct  5 10:42:32 1995 by James Miller
 * Last Modified: Fri Nov  3 10:52:49 1995 by James Miller
 *
 * A 3D vector. I originally tried to separate the semantics between a
 * 3D vector and a 3D cartesian point.  For instance, certain operations
 * like dot products are not defined on points.  However, if you consider
 * a point as a vector from the origin, then these operations can be
 * considered defined.  Treating vectors and points as vectors simply
 * makes the vector algebra more efficient and cleaner.
 *
 * Note that if we used projective coordinates, then vectors and points
 * could be explicitly distinguished.
 *
 ************************************************************************/

#ident "@(#)Vector3.h	1.1 95/11/09 RPI"

#include <math.h>
#include <stream.h>

class Vector3 {
  public:	
    // constructors
    Vector3()  { xval = 0.0; yval = 0.0; zval = 0.0; }
    Vector3(double vx, double vy, double vz)
               { xval = vx; yval = vy; zval = vz; }
    Vector3(const Vector3& vec)
               { xval = vec.xval; yval = vec.yval; zval = vec.zval; }

    // read access routines (can be called from constant Vector3's)
    double x() const { return xval; }
    double y() const { return yval; }
    double z() const { return zval; }

    double  norm() { return sqrt(xval*xval + yval*yval + zval*zval); }

    // write access routines
    Vector3& x( double vx ) { xval = vx; return *this; }
    Vector3& y( double vy ) { yval = vy; return *this; }
    Vector3& z( double vz ) { zval = vz; return *this; }


    // vector operators
    Vector3& operator = (const Vector3& vec);
    Vector3& operator +=(const Vector3& vec);
    Vector3& operator -=(const Vector3& vec);

    Vector3 operator + (const Vector3& vec) const;  
    Vector3 operator - (const Vector3& vec) const;  
    Vector3 operator - () const;                      // negation

    Vector3 operator % (const Vector3& vec) const;    // cross product
    double  operator * (const Vector3& vec) const;    // dot product

    // scalar operators
    Vector3& operator *= (double scale);
    Vector3& operator /= (double scale);
    
    Vector3 operator * (double scale) const;
    Vector3 operator / (double scale) const;
    
    
    // destructor
    virtual ~Vector3();	


  protected:
    double xval, yval, zval;

};

inline Vector3& Vector3::operator=(const Vector3& vec)
{
    xval = vec.xval; yval = vec.yval; zval = vec.zval;
    return *this;
}

inline Vector3& Vector3::operator+=(const Vector3& vec)
{
    xval += vec.xval; yval += vec.yval; zval += vec.zval;
    return *this;
}

inline Vector3& Vector3::operator-=(const Vector3& vec)
{
    xval -= vec.xval; yval -= vec.yval; zval -= vec.zval;
    return *this;
}


// dot product
inline double Vector3::operator*(const Vector3& vec) const
{
    return (xval*vec.xval + yval*vec.yval + zval*vec.zval);
}


//
// scalar operators...
//


inline Vector3& Vector3::operator*=(double scale)
{
    xval *= scale;  yval *= scale;  zval *= scale;

    return *this;
}

inline Vector3& Vector3::operator/=(double scale)
{
    xval /= scale;  yval /= scale;  zval /= scale;

    return *this;
}


#endif /* _VECTOR3_H */
