
#ifndef _WINDOWSTAT_H
#define _WINDOWSTAT_H

/************************************************************************
 *
 * File: WindowStat.h
 *
 * Copyright (c) 1995 by Rensselaer Polytechnic Institute,
 *                       Troy, NY 12180
 *
 * Created:       Wed Oct 11 09:53:28 1995 by James Miller
 * Last Modified: Mon Jan 29 10:40:58 1996 by James Miller
 *
 * A template class to hold a "window" or 2D array of some class.
 * The purpose of this class is to be able to hold a window of data
 * and/or (per pixel) statistics on that data.
 *
 * The storage is implemented as a 1D array of pointers to the stored
 * class.  This allows for the array subscripting [i][j] to return
 * a pointer to the stored object at [i][j] and therefore avoid
 * "copying" the class onto and off from the stack.
 *
 * Note that this class is responsible for its own memory allocation/
 * deallocation.  It allocates the array of pointers AND each instance
 * of the stored type.  To place information into the window, the
 * data can be "copied" into the reference returned through [][].
 * Note that any reference used via [][] should never be freed
 * outside of the WindowStat destructor.
 *
 * Finally, note that WindowStat allocates its own memory, the stored
 * class at each pixel must have a default (parameterless) constructor
 * and a constructor that can take in a default value (const Type &def).
 *
 ************************************************************************/

#ident "@(#)WindowStat.h	1.4 96/01/29 RPI"

#include <assert.h>

template <class Type>
class WindowStat {
  public:	
    WindowStat(int rows, int cols);	
    WindowStat(int rows, int cols, const Type &default_value);	
    virtual ~WindowStat();	

    // a_windowstat[row][col] will return a pointer to a Type
    Type** operator[](int row);
    const Type *const * operator[](int row) const;

    // get the size of the window
    int rows() const { return num_rows; };
    int cols() const { return num_cols; };
    
  protected:
    // routines to allocate/deallocate the window of pointers and the Type's
    void allocate();
    void allocate(const Type &default_value);
    void deallocate();
    
    Type **win;                       // 2D array is held as 1D array

    int num_rows, num_cols;          // size of the window

};


template<class Type>
WindowStat<Type>::WindowStat(int rows, int cols)
{
    assert((rows > 0) && (cols > 0));
    
    num_rows = rows;
    num_cols = cols;

    allocate();
}


// constructor that allows for a default data value to be used
template<class Type>
WindowStat<Type>::WindowStat(int rows, int cols, const Type &default_value)
{
    assert((rows > 0) && (cols > 0));
    
    num_rows = rows;
    num_cols = cols;

    allocate(default_value);
}


template<class Type>
WindowStat<Type>::~WindowStat()
{
    deallocate();
}


// return a pointer to the specified "row" of the window...
template<class Type>
Type** WindowStat<Type>::operator[] (int row)
{
    assert((row >= 0) && (row < num_rows));

    return (win + row*num_cols);
}

// read access routine for constant WindowStat objects
template<class Type>
const Type *const * WindowStat<Type>::operator[](int row) const
{
    assert((row >= 0) && (row < num_rows));

    return ((const Type *const *) (win + row*num_cols));
}


//
// Protected routine to allocate the pointer window and each of the
// instances of Type
//
template<class Type>
void WindowStat<Type>::allocate()
{
    // allocate the array of pointers
    win = new Type*[num_rows*num_cols];
    assert(win != 0);

    // allocate each of the Type's
    for (int i=0; i < num_rows*num_cols; i++)
    {
	win[i] = new Type;
	assert(win[i] != 0);
    }
    
}

// allocate instances of Type that match a default parameter
template<class Type>
void WindowStat<Type>::allocate(const Type &default_value)
{
    // allocate the array of pointers
    win = new Type*[num_rows*num_cols];
    assert(win != 0);

    // allocate each of the Type's
    for (int i=0; i < num_rows*num_cols; i++)
    {
	win[i] = new Type(default_value);
	assert(win[i] != 0);
    }
    
}


//
// Protected routine to deallocate the pointer window and each of the
// instances of Type
//
template<class Type>
void WindowStat<Type>::deallocate()
{
    // delete each of the Type's allocated
    for (int i=0; i < num_rows*num_cols; i++)
	delete win[i];

    // free up the memory to hold the pointers
    delete [] win;
}
#endif /* _WINDOWSTAT_H */
