
/************************************************************************
 *
 * File: misc.cc
 *
 * Copyright (c) 1995 by Rensselaer Polytechnic Institute,
 *                       Troy, NY 12180
 *
 * Created:       Thu Oct  5 11:10:15 1995 by James Miller
 * Last Modified: Tue Apr  2 11:23:46 1996 by James Miller
 *
 * Miscellaneous matrix and math routines, the majority of which are
 * from "Numerical Recipes in C: The Art of Scientific Computing"
 *
 * gaussian_cdf_inv() is an adaptation of a routine provided by Robert W.
 * Cox from the Biophysics Research Institute at the Medical College of
 * Wisconsin.
 *
 ************************************************************************/

#ident "@(#)misc.cc	1.5 96/04/04 RPI"

#include <math.h>
#include <stream.h>
#include <stdlib.h>

#include "misc.h"


/************************************************************************ 
 ***************   Numerical Recipes Routines   *************************
 ************************************************************************/ 



double erf1( double x)
{
    return x < 0.0 ? -gammp(0.5,x*x) : gammp(0.5,x*x);
}

double erfc( double x)
{
    return x < 0.0 ? 1.0+gammp(0.5,x*x) : gammq(0.5,x*x);
}

// Chebychev approximation to erfc
double erfcc( double x )
{
    double t,z,ans;
 
    z=fabs(x);
    t=1.0/(1.0+0.5*z);
    ans=t*exp(-z*z-1.26551223+t*(1.00002368+t*(0.37409196+t*(0.09678418+
	   t*(-0.18628806+t*(0.27886807+t*(-1.13520398+t*(1.48851587+
	         t*(-0.82215223+t*0.17087277)))))))));
    return  x >= 0.0 ? ans : 2.0-ans;
}


double gammln( double xx )
{
    double x,tmp,ser;
    static double cof[6]={76.18009173,-86.50532033,24.01409822,
                          -1.231739516,0.120858003e-2,-0.536382e-5};
    int j;

    x=xx-1.0;
    tmp=x+5.5;
    tmp -= (x+0.5)*log(tmp);
    ser=1.0;
    for (j=0;j<=5;j++) {
        x += 1.0;
        ser += cof[j]/x;
    }
    return -tmp+log(2.50662827465*ser);
}

double gammp( double a, double x)
{
    double gamser,gammcf,gln;

    if (x < 0.0 || a <= 0.0) nrerror("Invalid arguments in routine GAMMP");
    if (x < (a+1.0)) {
        gser(&gamser,a,x,&gln);
        return gamser;
    } else {
        gcf(&gammcf,a,x,&gln);
        return 1.0-gammcf;
    }
}

double gammq( double a, double x )
{
    double gamser,gammcf,gln;

    if (x < 0.0 || a <= 0.0) nrerror("Invalid arguments in routine GAMMQ");
    if (x < (a+1.0)) {
      gser(&gamser,a,x,&gln);
      return 1.0-gamser;
    } else {
      gcf(&gammcf,a,x,&gln);
      return gammcf;
    }
}



#define ITMAX 100
#define EPS 3.0e-7

void gcf( double * gammcf, double a, double x, double * gln)
{
    int n;
    float gold=0.0,g,fac=1.0,b1=1.0;
    float b0=0.0,anf,ana,an,a1,a0=1.0;

    *gln=gammln(a);
    a1=x;
    for (n=1;n<=ITMAX;n++) {
        an=(double) n;
        ana=an-a;
        a0=(a1+a0*ana)*fac;
        b0=(b1+b0*ana)*fac;
        anf=an*fac;
        a1=x*a0+anf*a1;
        b1=x*b0+anf*b1;
        if (a1) {
            fac=1.0/a1;
            g=b1*fac;
            if (fabs((g-gold)/g) < EPS) {
                *gammcf=exp(-x+a*log(x)-(*gln))*g;
                return;
            }
            gold=g;
        }
    }
    nrerror("a too large, ITMAX too small in routine GCF");
}

#undef ITMAX
#undef EPS

#define ITMAX 100
#define EPS 3.0e-7

void gser( double * gamser, double a, double x, double * gln)
{
    int n;
    double sum,del,ap;

    *gln=gammln(a);
    if (x <= 0.0) {
        if (x < 0.0) nrerror("x less than 0 in routine GSER");
        *gamser=0.0;
        return;
    } else {
        ap=a;
        del=sum=1.0/a;
        for (n=1;n<=ITMAX;n++) {
            ap += 1.0;
            del *= x/ap;
            sum += del;
            if (fabs(del) < fabs(sum)*EPS) {
                *gamser=sum*exp(-x+a*log(x)-(*gln));
                return;
            }
        }
        nrerror("a too large, ITMAX too small in routine GSER");
        return;
    }
}

#undef ITMAX
#undef EPS



double betai(double a, double b, double x)
{
        double bt;

        if (x < 0.0 || x > 1.0) nrerror("Bad x in routine BETAI");
        if (x == 0.0 || x == 1.0) bt=0.0;
        else
                bt=exp(gammln(a+b)-gammln(a)-gammln(b)+a*log(x)+b*log(1.0-x));
        if (x < (a+1.0)/(a+b+2.0))
                return bt*betacf(a,b,x)/a;
        else
                return 1.0-bt*betacf(b,a,1.0-x)/b;
}



#define ITMAX 100
#define EPS 3.0e-7

double betacf(double a, double b, double x)
{
        double qap,qam,qab,em,tem,d;
        double bz,bm=1.0,bp,bpp;
        double az=1.0,am=1.0,ap,app,aold;
        int m;

        qab=a+b;
        qap=a+1.0;
        qam=a-1.0;
        bz=1.0-qab*x/qap;
        for (m=1;m<=ITMAX;m++) {
                em=(double) m;
                tem=em+em;
                d=em*(b-em)*x/((qam+tem)*(a+tem));
                ap=az+d*am;
                bp=bz+d*bm;
                d = -(a+em)*(qab+em)*x/((qap+tem)*(a+tem));
                app=ap+d*az;
                bpp=bp+d*bz;
                aold=az;
                am=ap/bpp;
                bm=bp/bpp;
                az=app/bpp;
                bz=1.0;
                if (fabs(az-aold) < (EPS*fabs(az))) return az;
        }
        nrerror("a or b too big, or ITMAX too small in BETACF");

	return 0.0;   // to avoid compiler warnings
}

#undef ITMAX
#undef EPS



void nrerror( char error_text[] )
{
    cerr << "Numerical Recipes run-time error... " << endl;
    cerr << error_text << endl;
    cerr << "...now exiting to system... " << endl;
    exit(1);
}



float *fvector(int nl, int nh)
{
        float *v;

        v=(float *)malloc((unsigned) (nh-nl+1)*sizeof(float));
        if (!v) nrerror("allocation failure in vector()");
        return v-nl;
}


int *ivector(int nl, int nh)
{
        int *v;

        v=(int *)malloc((unsigned) (nh-nl+1)*sizeof(int));
        if (!v) nrerror("allocation failure in ivector()");
        return v-nl;
}

double *dvector(int nl, int nh)
{
        double *v;

        v=(double *)malloc((unsigned) (nh-nl+1)*sizeof(double));
        if (!v) nrerror("allocation failure in dvector()");
        return v-nl;
}


float **fmatrix(int nrl, int nrh, int ncl, int nch)
{
        int i;
        float **m;

        m=(float **) malloc((unsigned) (nrh-nrl+1)*sizeof(float*));
        if (!m) nrerror("allocation failure 1 in matrix()");
        m -= nrl;

        for(i=nrl;i<=nrh;i++) {
                m[i]=(float *) malloc((unsigned) (nch-ncl+1)*sizeof(float));
                if (!m[i]) nrerror("allocation failure 2 in matrix()");
                m[i] -= ncl;
        }
        return m;
}

double **dmatrix(int nrl, int nrh, int ncl, int nch)
{
        int i;
        double **m;

        m=(double **) malloc((unsigned) (nrh-nrl+1)*sizeof(double*));
        if (!m) nrerror("allocation failure 1 in dmatrix()");
        m -= nrl;

        for(i=nrl;i<=nrh;i++) {
                m[i]=(double *) malloc((unsigned) (nch-ncl+1)*sizeof(double));
                if (!m[i]) nrerror("allocation failure 2 in dmatrix()");
                m[i] -= ncl;
        }
        return m;
}


//
// These routine added so that a vector or matrix could be re-allocated.
//   note that realloc does not provide a means to determine if a resize was
//   successful
//

float *re_fvector(float *vec, int old_nl, int old_nh, int new_nl, int new_nh)
{
        float *v = vec;

        v=(float *)realloc(v+old_nl,(unsigned)(new_nh-new_nl+1)*sizeof(float));
        return v-new_nl;
}


int *re_ivector(int *vec, int old_nl, int old_nh, int new_nl, int new_nh)
{
        int *v = vec;

        v=(int *)realloc(v+old_nl, (unsigned) (new_nh-new_nl+1)*sizeof(int));
        return v-new_nl;
}

double *re_dvector(double *vec, int old_nl, int old_nh, int new_nl, int new_nh)
{
        double *v = vec;

        v=(double *)realloc(v+old_nl,
			    (unsigned) (new_nh-new_nl+1)*sizeof(double));
        return v-new_nl;
}


float **re_fmatrix(float **mat,
		   int old_nrl, int old_nrh, int old_ncl, int old_nch,
		   int new_nrl, int new_nrh, int new_ncl, int new_nch)
{
        int i;
        float **m = mat;

	if (old_nrl != new_nrl)
	  nrerror("First row index for new and old matrix must match.");
	
	// extend old lengths of old rows
	for (i=old_nrl; i<=old_nrh; i++) {
	  m[i] = (float *)realloc(m[i]+old_ncl,
			 (unsigned) (new_nch - new_ncl + 1)*sizeof(float));
	  m[i] -= new_ncl;
	}
	  

	// reallocate the row pointers
        m=(float **) realloc(m+old_nrl,
			     (unsigned) (new_nrh-new_nrl+1)*sizeof(float*));
        m -= new_nrl;


	// allocate the new rows
        for(i=old_nrh+1;i<=new_nrh;i++) {
	  m[i]=(float *) malloc((unsigned) (new_nch-new_ncl+1)*sizeof(float));
	  if (!m[i]) nrerror("allocation failure 2 in matrix()");
	  m[i] -= new_ncl;
        }
        return m;
}

double **re_dmatrix(double **mat,
		    int old_nrl, int old_nrh, int old_ncl, int old_nch,
		    int new_nrl, int new_nrh, int new_ncl, int new_nch)
{
        int i;
        double **m = mat;

	if (old_nrl != new_nrl)
	  nrerror("First row index for new and old matrix must match.");
	
	// extend old lengths of old rows
	for (i=old_nrl; i<=old_nrh; i++) {
	  m[i] = (double *)realloc(m[i]+old_ncl,
			 (unsigned) (new_nch - new_ncl + 1)*sizeof(double));
	  m[i] -= new_ncl;
	}
	  

	// reallocate the row pointers
        m=(double **) realloc(m+old_nrl,
			     (unsigned) (new_nrh-new_nrl+1)*sizeof(double*));
        m -= new_nrl;


	// allocate the new rows
        for(i=old_nrh+1;i<=new_nrh;i++) {
	  m[i]=(double *)malloc((unsigned) (new_nch-new_ncl+1)*sizeof(double));
	  if (!m[i]) nrerror("allocation failure 2 in matrix()");
	  m[i] -= new_ncl;
        }
        return m;
}


//
// end of reallocation routines
//


void free_ivector(int *v, int nl, int nh)
{
        free((char*) (v+nl));
}

void free_dvector(double *v, int nl, int nh)
{
        free((char*) (v+nl));
}

void free_fvector(float *v, int nl, int nh)
{
        free((char*) (v+nl));
}

void free_fmatrix(float **m, int nrl, int nrh, int ncl, int nch)
{
        int i;

        for(i=nrh;i>=nrl;i--) free((char*) (m[i]+ncl));
        free((char*) (m+nrl));
}

void free_dmatrix(double **m, int nrl, int nrh, int ncl, int nch)
{
        int i;

        for(i=nrh;i>=nrl;i--) free((char*) (m[i]+ncl));
        free((char*) (m+nrl));
}


#define TINY 1.0e-20;

void ludcmp(double **a, int n, int *indx, double *d)
{
        int i,imax=1,j,k;
        double big,dum,sum,temp;
        double *vv;
//	  float *vector();
//        void nrerror(),free_vector();

        vv=dvector(1,n);
        *d=1.0;
        for (i=1;i<=n;i++) {
                big=0.0;
                for (j=1;j<=n;j++)
                        if ((temp=fabs(a[i][j])) > big) big=temp;
                if (big == 0.0) nrerror("Singular matrix in routine LUDCMP");
                vv[i]=1.0/big;
        }
        for (j=1;j<=n;j++) {
                for (i=1;i<j;i++) {
                        sum=a[i][j];
                        for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
                        a[i][j]=sum;
                }
                big=0.0;
                for (i=j;i<=n;i++) {
                        sum=a[i][j];
                        for (k=1;k<j;k++)
                                sum -= a[i][k]*a[k][j];
                        a[i][j]=sum;
                        if ( (dum=vv[i]*fabs(sum)) >= big) {
                                big=dum;
                                imax=i;
                        }
                }
                if (j != imax) {
                        for (k=1;k<=n;k++) {
                                dum=a[imax][k];
                                a[imax][k]=a[j][k];
                                a[j][k]=dum;
                        }
                        *d = -(*d);
                        vv[imax]=vv[j];
                }
                indx[j]=imax;
                if (a[j][j] == 0.0) a[j][j]=TINY;
                if (j != n) {
                        dum=1.0/(a[j][j]);
                        for (i=j+1;i<=n;i++) a[i][j] *= dum;
                }
        }
        free_dvector(vv,1,n);
}

#undef TINY
 

void lubksb(double **a, int n, int *indx, double b[])
{
        int i,ii=0,ip,j;
        double sum;

        for (i=1;i<=n;i++) {
                ip=indx[i];
                sum=b[ip];
                b[ip]=b[i];
                if (ii)
                        for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
                else if (sum) ii=i;
                b[i]=sum;
        }
        for (i=n;i>=1;i--) {
                sum=b[i];
                for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
                b[i]=sum/a[i][i];
        }
}



//
// Singular Value Decomposition routines
//

static double at,bt,ct;
#define PYTHAG(a,b) ((at=fabs(a)) > (bt=fabs(b)) ? \
(ct=bt/at,at*sqrt(1.0+ct*ct)) : (bt ? (ct=at/bt,bt*sqrt(1.0+ct*ct)): 0.0))

static double maxarg1,maxarg2;
#define MAX(a,b) (maxarg1=(a),maxarg2=(b),(maxarg1) > (maxarg2) ?\
	(maxarg1) : (maxarg2))
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

//
// This routine was altered such that the vector allocated is only allocated
// if the space requirements exceed the previous space requirements.
// (Note we overallocate.)
//
// We have to rely on the memory manager to free this memory upon exiting.
//
void svdcmp(double **a, int m, int n, double *w, double **v)
{
    int flag,i,its,j,jj,k,l=0,nm=0;
    double c,f,h,s,x,y,z;
    double anorm=0.0,g=0.0,scale=0.0;
    static double *rv1 = NULL;
    static int old_n = 0;
    
//    float *dvector();
//    void nrerror(),free_dvector();

    if (m < n) nrerror("SVDCMP: You must augment A with extra zero rows");
    // original line ...
    //    rv1=dvector(1,n);
    //
    // changed to:
    if (rv1 == NULL) {
      rv1 = dvector(1, n);
      old_n = n;
    }
    else if (n > old_n) {
      rv1 = re_dvector(rv1, 1, old_n, 1, 2*n);
      old_n = 2*n;
    }
    
    for (i=1;i<=n;i++) {
	l=i+1;
	rv1[i]=scale*g;
	g=s=scale=0.0;
	if (i <= m) {
	    for (k=i;k<=m;k++) scale += fabs(a[k][i]);
	    if (scale) {
		for (k=i;k<=m;k++) {
		    a[k][i] /= scale;
		    s += a[k][i]*a[k][i];
		}
		f=a[i][i];
		g = -SIGN(sqrt(s),f);
		h=f*g-s;
		a[i][i]=f-g;
		if (i != n) {
		    for (j=l;j<=n;j++) {
			for (s=0.0,k=i;k<=m;k++) s += a[k][i]*a[k][j];
			f=s/h;
			for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
		    }
		}
		for (k=i;k<=m;k++) a[k][i] *= scale;
	    }
	}
	w[i]=scale*g;
	g=s=scale=0.0;
	if (i <= m && i != n) {
	    for (k=l;k<=n;k++) scale += fabs(a[i][k]);
	    if (scale) {
		for (k=l;k<=n;k++) {
		    a[i][k] /= scale;
		    s += a[i][k]*a[i][k];
		}
		f=a[i][l];
		g = -SIGN(sqrt(s),f);
		h=f*g-s;
		a[i][l]=f-g;
		for (k=l;k<=n;k++) rv1[k]=a[i][k]/h;
		if (i != m) {
		    for (j=l;j<=m;j++) {
			for (s=0.0,k=l;k<=n;k++) s += a[j][k]*a[i][k];
			for (k=l;k<=n;k++) a[j][k] += s*rv1[k];
		    }
		}
		for (k=l;k<=n;k++) a[i][k] *= scale;
	    }
	}
	anorm=MAX(anorm,(fabs(w[i])+fabs(rv1[i])));
    }
    for (i=n;i>=1;i--) {
	if (i < n) {
	    if (g) {
		for (j=l;j<=n;j++)
		    v[j][i]=(a[i][j]/a[i][l])/g;
		for (j=l;j<=n;j++) {
		    for (s=0.0,k=l;k<=n;k++) s += a[i][k]*v[k][j];
		    for (k=l;k<=n;k++) v[k][j] += s*v[k][i];
		}
	    }
	    for (j=l;j<=n;j++) v[i][j]=v[j][i]=0.0;
	}
	v[i][i]=1.0;
	g=rv1[i];
	l=i;
    }
    for (i=n;i>=1;i--) {
	l=i+1;
	g=w[i];
	if (i < n)
	    for (j=l;j<=n;j++) a[i][j]=0.0;
	if (g) {
	    g=1.0/g;
	    if (i != n) {
		for (j=l;j<=n;j++) {
		    for (s=0.0,k=l;k<=m;k++) s += a[k][i]*a[k][j];
		    f=(s/a[i][i])*g;
		    for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
		}
	    }
	    for (j=i;j<=m;j++) a[j][i] *= g;
	} else {
	    for (j=i;j<=m;j++) a[j][i]=0.0;
	}
	++a[i][i];
    }
    for (k=n;k>=1;k--) {
	for (its=1;its<=30;its++) {
	    flag=1;
	    for (l=k;l>=1;l--) {
		nm=l-1;
		if (fabs(rv1[l])+anorm == anorm) {
		    flag=0;
		    break;
		}
		if (fabs(w[nm])+anorm == anorm) break;
	    }
	    if (flag) {
		c=0.0;
		s=1.0;
		for (i=l;i<=k;i++) {
		    f=s*rv1[i];
		    if (fabs(f)+anorm != anorm) {
			g=w[i];
			h=PYTHAG(f,g);
			w[i]=h;
			h=1.0/h;
			c=g*h;
			s=(-f*h);
			for (j=1;j<=m;j++) {
			    y=a[j][nm];
			    z=a[j][i];
			    a[j][nm]=y*c+z*s;
			    a[j][i]=z*c-y*s;
			}
		    }
		}
	    }
	    z=w[k];
	    if (l == k) {
		if (z < 0.0) {
		    w[k] = -z;
		    for (j=1;j<=n;j++) v[j][k]=(-v[j][k]);
		}
		break;
	    }
	    if (its == 30) nrerror("No convergence in 30 SVDCMP iterations");
	    x=w[l];
	    nm=k-1;
	    y=w[nm];
	    g=rv1[nm];
	    h=rv1[k];
	    f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
	    g=PYTHAG(f,1.0);
	    f=((x-z)*(x+z)+h*((y/(f+SIGN(g,f)))-h))/x;
	    c=s=1.0;
	    for (j=l;j<=nm;j++) {
		i=j+1;
		g=rv1[i];
		y=w[i];
		h=s*g;
		g=c*g;
		z=PYTHAG(f,h);
		rv1[j]=z;
		c=f/z;
		s=h/z;
		f=x*c+g*s;
		g=g*c-x*s;
		h=y*s;
		y=y*c;
		for (jj=1;jj<=n;jj++) {
		    x=v[jj][j];
		    z=v[jj][i];
		    v[jj][j]=x*c+z*s;
		    v[jj][i]=z*c-x*s;
		}
		z=PYTHAG(f,h);
		w[j]=z;
		if (z) {
		    z=1.0/z;
		    c=f*z;
		    s=h*z;
		}
		f=(c*g)+(s*y);
		x=(c*y)-(s*g);
		for (jj=1;jj<=m;jj++) {
		    y=a[jj][j];
		    z=a[jj][i];
		    a[jj][j]=y*c+z*s;
		    a[jj][i]=z*c-y*s;
		}
	    }
	    rv1[l]=0.0;
	    rv1[k]=f;
	    w[k]=x;
	}
    }
    //free_dvector(rv1,1,n);
}

#undef SIGN
#undef MAX
#undef PYTHAG

//
// This routine was altered such that the vector allocated is only allocated
// if the space requirements exceed the previous space requirements.
// (Note we overallocate.)
//
// We have to rely on the memory manager to free this memory upon exiting.
//
void svbksb(double **u, double *w, double **v, int m, int n,
	    double *b, double *x)
{
        int jj,j,i;
        double s;
	static double *tmp = NULL;
	static int old_n = 0;
//	  double *dvector();
//        void free_dvector();

    // original line ...
    //    tmp=dvector(1,n);
    //
    // changed to:
    if (tmp == NULL) {
      tmp = dvector(1, n);
      old_n = n;
    }
    else if (n > old_n) {
      tmp = re_dvector(tmp, 1, old_n, 1, 2*n);
      old_n = 2*n;
    }

    for (j=1;j<=n;j++) {
      s=0.0;
      if (w[j]) {
	for (i=1;i<=m;i++) s += u[i][j]*b[i];
	s /= w[j];
      }
      tmp[j]=s;
    }
    for (j=1;j<=n;j++) {
      s=0.0;
      for (jj=1;jj<=n;jj++) s += v[j][jj]*tmp[jj];
      x[j]=s;
    }
    // free_dvector(tmp,1,n);
}


/*
  END OF NUMERICAL RECIPES IN C CODE
*/  




//
// Robert W. Cox from the Biophysics Research Institute at the Medical
// College of Wisconsin provided the following routine for computing
// the inverse CDF for a standardized Gaussian.  This function is
// based off of a rational polynomial approximation to the inverse
// Gaussian CDF which can be found in
//
// M. Abramowitz and I.A. Stegun. Handbook of Mathematical Functions
// with Formulas, Graphs, and Mathematical Tables.  John Wiley & Sons.
// New York. Equation 26.2.23. pg. 933. 1972.
//
// Since the initial approximation only provides an estimate within
// 4.5 E-4 of the true value, 3 Newton-Raphson interations are used
// to refine the approximation.
//
// Let,
// Q(x) = (1/sqrt(2*pi)) Int_{x}^{infinity} e^{-t^2/2} dt
//      = 0.5 * erfc(x/sqrt(2))
//
// Given p, this function computes x such that Q(x) = p, for 0 < p < 1
//
// Note that the Gaussian CDF is defined as
// P(x) = (1/sqrt(2*pi)) Int_{-infinity}{x} e^{-t^2/2} dt
//      = 1 - Q(x)
//
// This function has been modified to compute the inverse of P(x) instead
// of Q(x).
//
double gaussian_cdf_inv( double p )
{
   double dp , dx , dt , ddq , dq ;
   int    newt ;

   dp = (p <= 0.5) ? (p) : (1.0-p) ;   /* make between 0 and 0.5 */

   // if original value is invalid, return +infinity or -infinity
   // changed from original code to reflect the fact that the
   // the inverse of P(x) not Q(x) is desired.
   //
   // Original line: used for inverse of Q(x)
   // if( dp <= 0.0 ){ dx = 13.0 ;  return ( (p <= 0.5) ? (dx) : (-dx) ) ; }

   // replaced with this if construct for the inverse of P(x)
   if (p <= 0.0)
     return -HUGE_VAL;
   else if (p >= 1.0)
     return HUGE_VAL;

   
   /**  Step 1:  use 26.2.23 from Abramowitz and Stegun **/

   dt = sqrt( -2.0 * log(dp) ) ;
   dx = dt
     - ((.010328e+0*dt + .802853e+0)*dt + 2.515517e+0)
     /(((.001308e+0*dt + .189269e+0)*dt + 1.432788e+0)*dt + 1.e+0) ;

   /**  Step 2:  do 3 Newton steps to improve this **/

   for( newt=0 ; newt < 3 ; newt++ ){
     dq  = 0.5e+0 * erfcc( dx / 1.414213562373095e+0 ) - dp ;
     ddq = exp( -0.5e+0 * dx * dx ) / 2.506628274631000e+0 ;
     dx  = dx + dq / ddq ;
   }

   // original line when computing the inverse of Q(x)
   // return ( (p <= 0.5) ? (dx) : (-dx) ) ;  /* return with correct sign */
   //
   // Note that P(-x) = Q(x), so whatever x was calculated for Q(x) = p,
   // we simply need to return the negative of x to get P(xp) = p.
   //
   return ( (p <= 0.5) ? (-dx) : (dx) ); // return with correct sign
}

