

/************************************************************************
 *
 * File: muse.cc
 *
 * Copyright (c) 1995 by Rensselaer Polytechnic Institute,
 *                       Troy, NY 12180
 *
 * Created:       Thu Oct  5 10:11:42 1995 by James Miller
 * Last Modified: Mon Apr  8 11:37:03 1996 by James Miller
 *
 *
 * MUSE:  Minimum Unbiased Scale Estimates
 *
 * MUSE is a robust local surface patch extractor.  An input range
 * scene is divided into a series of non-overlapping windows and a
 * set of surfaces is extracted from each window.  A reconstructed range
 * image, a surface description file, and a segmentation image are provided
 * on output.
 *
 * This program implements the operator described in:
 *
 *     Miller, J.V., Stewart, C.V. ``MUSE: Robust Surface Fitting using
 *       Unbiased Scale Estimates''.  To appear in Computer Vision and
 *       Pattern Recognition 1996.
 *
 * which can be found at:
 *
 *     http://www.cs.rpi.edu/~millerj/papers/muse_cvpr96.ps.gz
 *
 * Usage:
 *
 *     muse parameter_file in_range_file out_range_file out_surface_file
 *          [out_segment_file]
 *
 * Inputs:
 *
 *     parameter_file:
 *          Contains the parameters used to control MUSE.  These include
 *          the window size, which version of MUSE to use (original or
 *          interior point selection), values used to determine the number
 *          of random samples needed, and confidence thresholds. 
 *
 *          Available parameters to set:
 *             MU             -- Multiplier of estimated sigma used to
 *                               determine inliers
 *             MU_INT         -- Multiplier of estimated sigma used to
 *                               determine inliers interior to a random sample
 *             IGNORE_PCT     -- Algorithm ignores the first IGNORE_PCT
 *                               residuals in evaluating the min sk value
 *             WINDOW_SIZE    -- Two integers indicating the number of rows
 *                               and cols in a window, window size if currently
 *                               limited by the E[min s | k] table to
 *                               win_rows*win_cols <= 249
 *             DATA_MISSING   -- Float indicating the value of invalid range
 *                               measurements
 *             INTERIOR       -- Should algorithm restrict extraction
 *                               evaluation to range pixels "interior" to a
 *                               random sample, true or false
 *             MIN_FEATURE    -- Number of range points needed to be consider
 *                               a valid surface.
 *             EXTRACT_NUMBER -- Number of surfaces to extract
 *
 *          The following parameters are used in determining the number of
 *          random samples
 *             MAX_OUTLIER_PCT     -- maximum expected outlier percentage
 *             WINDOW_MAX_SURFACES -- expected maximum number of surfaces in
 *                                    a window
 *             MIN_SAMPLES         -- minimum number of random samples to use
 *             PROB_GOOD_SAMPLE    -- probability that the sampling process
 *                                    must match for finding a sample from a
 *                                    single surface
 *
 *          Format:  
 *            Each parameter is listed separately on its own line.  The
 *            format is
 *              PARAMETER_NAME value
 *
 *            Lines beginning with a '#' are ignored.
 *            Parameters not set will be given a default value.
 *            Parameters can be specified in any order.
 *            See the example file muse.param
 *
 *
 *     in_range_file:
 *          A range image in the .cam format. These range images have a header
 *          of two integers indicating the number of rows and columns followed
 *          by either:
 *                (1) rows*columns floats for the X positions, rows*columns
 *                    floats for the Y positions, and rows*columns float
 *                    for the Z positions,   [This would be an xyz file.]
 *          or
 *                (2) rows*columns floats for the Z positions. [This would
 *                    be a just_z file.]
 *
 *          The type of range file (xyz or just_z) is deduced from the file
 *          size. 
 *
 *          Range images in this format typically have a range value reserved
 *          for invalid data.  The invalid data identifier can be set in the
 *          in the parameter_file.  Typically a value of -320 for the X, Y, and
 *          Z refers to an invalid measurement (i.e. a point that the sensor
 *          could not measure).
 *
 * Outputs:
 *      out_range_file:
 *          A range image in the .cam format containing the reconstructed
 *          positions calculated by MUSE.  Datapoints not assigned to any
 *          surface are marked as invalid in the range image.
 *
 *      out_surface_file:
 *          A file in IBM's DataExplorer format (typically with a .dx
 *          extension) containing a "mesh" of disjoint rectangles describing
 *          the extracted surfaces.  Each surface is represented as a
 *          rectangle which bounds the reconstructed positions for that
 *          surface's inliers.  Data is associated with each surface which
 *          encodes the estimated scale (standard deviation of the noise)
 *          for that surface patch.  Although IBM's DataExplorer can support
 *          binary data, the surface files are written in text format so
 *          that people without DataExplorer can easily extract the surface
 *          and sigma data. 
 *
 *     out_segment_file:
 *          An optional output file which contains a segmentation image for
 *          the range scene.  This is a binary file with a two integer header
 *          describing the rows and columns, followed by rows*columns
 *          unsigned integers for surface identifiers. A value of zero
 *          represents and unclassified pixel (i.e. surface identifiers
 *          start counting at one).
 *
 *
 *
 * Restrictions in this version:
 *
 * This version of MUSE is currently limited to extracting planar patches.
 * It uses a naive random sampling algorithm, a naive interior point
 * selection algorithm, and naive algorithms for determining whether a
 * window contains another surface and whether a surface contains enough
 * inliers to be considered a surface.
 *
 * Note that the maximum number of surfaces that will be extracted from
 * any region is defined by the parameter EXTRACT_NUMBER.
 *
 * Window sizes are currently limited by the provided E[min s | k] table
 * to contain less than 250 points.
 *
 * MUSE looks for the file E_s_k.table in the current working directory
 * only.
 *
 ************************************************************************/

#ident "@(#)muse.cc	1.6 96/04/08 RPI"

#ifndef bool
#include <bool.h>
#endif
#include <time.h>
#include <stream.h>
#include <vector.h>
#include <iterator.h>
#include <algo.h>
#include "Vector3.h"
#include "RangeImage.h"
#include "PixelStat.h"
#include "WindowStat.h"
#include "Ak.h"
#include "E_s_k.h"
#include "Surface.h"
#include "Plane.h"
#include "SurfaceStat.h"
#include "MuseParameters.h"


// extra data strutures

// typedefs to clarify the code


// forward declarations
bool SimpleDone(WindowStat<PixelStat> &, MuseParameters &, int, int, int);
bool SimpleSample(WindowStat<PixelStat> &, int, vector<PixelStat*> &);
bool ScanWindow(WindowStat<PixelStat> &, vector<PixelStat*> &,
		vector<PixelStat*> &);
bool ScanSample(WindowStat<PixelStat> &, vector<PixelStat*> &,
		vector<PixelStat*> &);
int  NumAvail(WindowStat<PixelStat> &);
Surface* DefineRandomSampleFit(vector<PixelStat*> &);
bool SimpleSurfaceSurvive(SurfaceStat &, MuseParameters &);
void FindMidMaxStdDev(double W[], int &i_mid, int &i_max);
void WriteDXHeader(ofstream &, vector<SurfaceStat*> &);
void WriteDXSurfaceDataConnectionsFields(ofstream &, vector<SurfaceStat*> &);
void WriteDXSurfaceVertices(ofstream &, Vector3 *);
void ComputeBoundingbox( Vector3 *boundingbox, vector<const Vector3*> *points);

// function objects

// comparison function object for sorting PixelStat pointers based on field r
struct ComparePixelStatPtrR
{
    int operator()(PixelStat* p1, PixelStat* p2)
                { return (p1->r() < p2->r()); }
};

// comparison function object for sorting PixelStat pointers based on field s
struct ComparePixelStatPtrS
{
    int operator()(PixelStat* p1, PixelStat* p2)
                { return (p1->s() < p2->s()); }
};

// function object to remove NULL SurfaceStat pointers from a list
struct RemoveNULLSurfaceStatPtr
{
    int operator()(SurfaceStat* s) { return ( s == NULL ); }
};

// Have the SurfaceStat* calculate a least squares fit to its point data
struct LSFitSurfaceStatPtr
{
    int operator()(SurfaceStat* s)
          {
	      s->sigma( s->surface()->ls_fit( *(s->points()) ) );
	      return true;
	  }
};

// Pull out the position field from each PixelStat
struct PixeltoPosition
{
    const Vector3* operator()(PixelStat* px)
               {
		   return px->position();
	       }
};
    



main(int argc, char** argv)
{
    MuseParameters muse_params;

    // parse the argument list, let's do this now before we discuss variables
    if ((argc < 5) || (argc > 6)) {
	cerr << endl
	     << "Usage: muse parameter_file in_range_file "
	     << "out_range_file out_surface_file [out_segment_file]"
	     << endl << endl;
	exit(0);
    }
    muse_params.read_parameters(argv[1]);
    // prepare the output surface file
    ofstream surfacefile(argv[4], ios::out);
    if (!surfacefile) {
	cerr << "Error: cannot open surface output file " << argv[4] << endl;
	cerr << "Exiting." << endl << endl;
	exit(0);
    }

    // output a file with pixels assigned to surfaces?
    bool output_segment = (argc == 6);
    ofstream segmentfile;
    if (output_segment)    {
	segmentfile.open(argv[5], ios::out);
	if (!segmentfile) {
	    cerr << "Error: cannot open segmentation output file "
		 << argv[5] << endl << "Exiting." << endl << endl;
	    exit(0);
	}
    }

	
    // storage for the algorithm...
    RangeImage ri;       // range images, used for both input and output
                                                   // Window to operate on
    WindowStat<PixelStat> win(muse_params.win_rows(),
			      muse_params.win_cols());  
    vector<PixelStat*>      pts;                    // Window as 1D structure
    vector<PixelStat*>      ran_sample;             // Holds the sample points
    vector<const Vector3*>  pts_for_int_fit;        // scratch vector
    vector<const Vector3*> *pts_for_fit;            // vector to allocate...
    vector<PixelStat*>     *pxs_for_fit;            // vector to allocate...
    Vector3                *bounding_box;           // vector to allocate...
    vector<SurfaceStat*>    win_surfaces, all_surfaces;
    vector<PixelStat*>      rescued_pxs;            
    
    double sigma_hat, opt_sigma;
    Surface *surf, *opt_surf;
    PixelStat* px_stat;
    SurfaceStat* sf_stat;
    Vector3 point;

    unsigned int *segment;                          // segmentation image
    
    // local iterators, etc.
    vector<PixelStat*>::iterator px_it;
    vector<SurfaceStat*>::iterator sf_it, opt_sf_it;
    int i, j, ii, jj, samp, num_pts, k0, numavail,window_numavail, num_added;
    unsigned int k;
    double d, threshold, opt_d;

    // Function hooks.
    //     Changing these function pointers will change how the algorithm
    //     is implemented.  There are hooks for:  determining whether a
    //     window is exhausted of surfaces, gathering a random sample,
    //     and for gathering the data points from the window.
    //
    // Usage:
    //     The WindowDoneHook could be based on simply the number of samples
    //     remaining or could use MINPRAN to deduce if there is structure
    //     remaining.
    //
    //     The SamplingHook is used to set the algorithm used to select
    //     random samples.
    //
    //     The ScanHook allows for the algorithm to evaluate a hypothesized
    //     fit from all the data points available in the window, or
    //     perhaps from the data points available in the window AND
    //     "interior" to the random sample.
    //
    //     The SurfaceSurviveHook is used to determine whether a given
    //     extracted surface has enough data (remaining) to be considered
    //     a nonrandom event.
    //
    bool (*WindowDoneHook)(WindowStat<PixelStat> &,
			   MuseParameters &,int,int,int) = SimpleDone;
    bool (*SamplingHook)(WindowStat<PixelStat> &, int, vector<PixelStat*> &)
	= SimpleSample;
    bool (*ScanHook)(WindowStat<PixelStat> &, vector<PixelStat*> &,
		     vector<PixelStat*> & ) = ScanWindow;
    bool (*SurfaceSurviveHook)(SurfaceStat &, MuseParameters &)
	= SimpleSurfaceSurvive;

    // tables of constants for the algorithm
    Ak ak;                                         // E[uk]
    E_s_k esk;                                     // E[s | k]

    // check to make sure that the requested window size is smaller than
    // the calculated table sizes
    if (muse_params.win_rows() * muse_params.win_cols() > esk.max_N())
    {
      cerr << "Error:  requested window size [" << muse_params.win_rows()
	<< ", " << muse_params.win_cols() << "] exceeds the current "
	<< "normalization table E[min s | k] (maximum N = "
	<< esk.max_N() << ")." << endl << "Exiting." << endl << endl;
      exit(0);
    }
    if (muse_params.win_rows() * muse_params.win_cols() > MAX_N)
    {
      cerr << "Error:  requested window size [" << muse_params.win_rows()
	<< ", " << muse_params.win_cols() << "] exceeds the current "
	<< "normalization table E[uk] (maximum N = "
	<< MAX_N << ")." << endl << "Exiting." << endl << endl;
      exit(0);
    }

    
    // set the appropriate scanning hook
    if (muse_params.is_interior())
	ScanHook = ScanSample;
    else
	ScanHook = ScanWindow;

    // set the seed for the random number generator
    time_t tval = time(NULL);
    srand((int)tval);
    
    // load in the range file
    ri.read(argv[2]);
    if (ri.get_type() == RangeImage::just_z)
	cerr << "This file is a just_z file." << endl;
    if (ri.get_type() == RangeImage::xyz)
	cerr << "This file is an xyz file." << endl;

    // allocate the segmentation image
    segment = new unsigned int [ri.rows()*ri.cols()];
    assert( segment != NULL );
    for (k=0; k <= ri.rows()*ri.cols(); k++)      // fill in default values
      segment[k] = 0;

    // for each window in the image (currently non-overlapping)
    for (i = 0; i + muse_params.win_rows() < ri.rows();
	 i += muse_params.win_rows())
	for (j = 0; j + muse_params.win_cols() < ri.cols();
	     j += muse_params.win_cols())
	{

	    // clear the window surface list
	    if (!win_surfaces.empty())
	      win_surfaces.erase(win_surfaces.begin(), win_surfaces.end());
	    if (!rescued_pxs.empty())
	      rescued_pxs.erase(rescued_pxs.begin(), rescued_pxs.end());
	    
	    // Copy the range values into the window and clear other fields
	    cerr << "Filling window [" << i << ", " << j << "]"
		 << "\tSurface count = " << all_surfaces.size()<< "\t\t"
		 << endl;
 
	    for (ii=0; ii < muse_params.win_rows(); ii++)
		for (jj=0; jj < muse_params.win_cols(); jj++)
		{
		    // grab the PixelStat pointer
		    px_stat = win[ii][jj];

		    // clear and set the appropriate fields
		    (*px_stat).clear().row(i+ii).col(j+jj)
		      .position(ri[i+ii][j+jj]);

		    // if the actual range data point is invalid, mark it
		    // as unavailable. (Saves checking later).
		    if ( !(ri.is_valid(i+ii, j+jj)) )
			px_stat->available(false);
		}

	    // calculate the number of available pts to fit
	    numavail = NumAvail(win);
	    window_numavail = numavail; // how many pts did we have to start
	    
	    // We will ignore the first k0-1 residuals
	    k0 = (int)floor(muse_params.ignore_pct() * numavail) + 1;
	    
	    // Loop until there are no more surfaces in the scene
	    while ( ! ((*WindowDoneHook)(win, muse_params,
					 win_surfaces.size(), k0, numavail)) )
	    {
		opt_sigma = HUGE_VAL;
		opt_surf = NULL;
		
		// at this point we know that there are at least
		// p + k0 points left, couple this with the check that
		// when using interior point selection, that each sample
		// contains enough interior data (at least k0), then
		// we do not need to check whether pts.begin() + k0 - 1
		// is within the bounds of the list

		// Draw the random samples and find the "best" surface
		for (samp = 0;
		     samp < muse_params.samples(win_surfaces.size(),
						 window_numavail, numavail);
		     samp++)
		{
		  // if using interior point selection, we need to make 
		  // sure a given sample has "enough" interior data, we can
		  // check this for both interior and not using pts.size()
		  do {
		    // Grab a random sample
		    (*SamplingHook)(win, muse_params.p(), ran_sample);

		    // Collect the available data points for testing
		    (*ScanHook)(win, ran_sample, pts);
		  } while ( pts.size() < k0 );
		    
		  // Define the surface through the random samples, this
		  // allocates the memory for the surface (currently planes)
		  surf = DefineRandomSampleFit(ran_sample);

		  // measure the absolute residuals of points to surface
		  for (px_it = pts.begin(); px_it != pts.end(); ++px_it)
		    (*px_it)->r(fabs(surf->distance(*((*px_it)->position()))));

		  // sort the residuals based on r()
		  sort(pts.begin(), pts.end(), ComparePixelStatPtrR());

		  // compute the sk values from the rk values
		  //   only need to compute from k0 on
		  num_pts = pts.size();
		  for (px_it = pts.begin() + k0 - 1, k=k0;
		       px_it != pts.end(); ++px_it, ++k)
		    (*px_it)->s( (*px_it)->r() / ak[k][num_pts] );

		  // find the minimum sk value, ignoring the first k0-1 pts
		  px_it = min_element(pts.begin() + k0 - 1, pts.end(),
				      ComparePixelStatPtrS());

		  // apply the compensation factor to correct for bias
		  // due to E[s | k], i.e. Exp value of s given min at k
		  sigma_hat = (*px_it)->s() /
		                  esk[px_it - pts.begin() + 1][pts.size()];

		  // If interior point selection, perform a LS fit to the
		  // interior samples within an inlier bound to refine the
		  // surface. Note that using Surface* and virtual functions,
		  // this code will work with any type of Surface subclass.
		  if (muse_params.is_interior())
		  {
		    // setup the points vector
		    if (!pts_for_int_fit.empty())
		      pts_for_int_fit.erase(pts_for_int_fit.begin(),
					    pts_for_int_fit.end());
			
		    // loop through pts finding samples within inlier bound
		    threshold = muse_params.mu_int()*sigma_hat;
		    for (px_it = pts.begin(); px_it != pts.end(); ++px_it)
		    {
		      d = fabs(surf->distance(*((*px_it)->position())));
		      if ( d <= threshold )
			pts_for_int_fit.push_back((*px_it)->position());
		    }

		    // add the random sample points to the pts_for_fit list
		    for (px_it = ran_sample.begin();
			 px_it != ran_sample.end();  ++px_it)
		      pts_for_int_fit.push_back( (*px_it)->position() );
			
		    // perform a LS fit on the pts list using the current
		    // surface's distance metric, note this changes surf
		    // (be careful other fitting routines in surface
		    // cache previous results)
		    sigma_hat = surf->ls_fit( pts_for_int_fit );
		  }

		  // if this is the minimum so far, save it
		  if (sigma_hat < opt_sigma)
		  {
		    // first delete the previous best surface
		    if (opt_surf != NULL)  delete opt_surf;

		    // store optimal parameters
		    opt_sigma = sigma_hat;
		    opt_surf = surf;
		  }
		  else delete surf;       // free up the surface memory
		    
		} // for samp

		
		// store opt_surf and opt_sigma in a surface list for window
		pts_for_fit = new vector<const Vector3*>;
		pxs_for_fit = new vector<PixelStat*>;
		bounding_box = new Vector3[4];
		sf_stat = new SurfaceStat(opt_surf, opt_sigma, pts_for_fit,
					  pxs_for_fit, bounding_box);
		win_surfaces.push_back(sf_stat);
		    

		// mark all potential inliers across the window as unavailable
		threshold = muse_params.mu()*opt_sigma;
		for (ii=0, num_added = 0; ii < win.rows(); ii++)
		  for (jj=0; jj < win.cols(); jj++)
		  {
		    // grab the PixelStat pointer
		    px_stat = win[ii][jj];
		    if (px_stat->available())
		    {
		      // calculate the residual distance
		      d=fabs(opt_surf->distance(*(px_stat->position())));

		      // if within potential inlier bound, mark it off
		      if (d <= threshold) {
			px_stat->available( false );
			num_added++;
		      }
		    }
		  }

		// Update the number of available points
		numavail -= num_added;

		// Update the number of points to ignore (k0-1)
		k0 = (int)floor(muse_params.ignore_pct() * numavail) + 1;
		
	    } // while

	    // done extracting surfaces from the window

	    // check to make sure that we did indeed find a surface in this
	    // window (line needed in case while() falls through on 1st pass
	    if (win_surfaces.empty())
		continue;                // no surfaces in window

	    // recalculate the estimated standard deviation for each
	    // surface by finding all points in the window that could
	    // potentially be inliers, and calculating the standard
	    // deviation of these residuals (assuming the mean is 0)
	    // Note that this allows points removed by an earlier surface in
	    // the window to contribute to the sigma_hat for a latter surface.
	    for( sf_it = win_surfaces.begin();
		 sf_it != win_surfaces.end(); ++sf_it)
	    {
	      sigma_hat = 0.0;
	      num_pts = 0;

	      // run through each pixel in the window
	      for (ii = 0; ii < win.rows(); ii++)
		for (jj = 0; jj < win.cols(); jj++)
		{
		  px_stat = win[ii][jj];

		  // is the point a valid pixel?
		  if ( ri.is_valid( px_stat->row(), px_stat->col() ))
		  {
		    // calculate the residual distance
		    d = fabs(((*sf_it)->surface())
			     ->distance(*(px_stat->position())));

		    // is it an inlier?
		    if (d <=  muse_params.mu() * (*sf_it)->sigma()) {
		      sigma_hat += (d*d);
		      num_pts++;
		    }
		  }
		}

		// calculate the new standard deviation, note that when
		// iterior selection is NOT used p of the residuals
		// used above will be exactly zero since they are the
		// the random sample points, therefore normalize by  num_pts-p
		if (muse_params.is_interior())
		  sigma_hat = sqrt( sigma_hat / (double)num_pts );
		else
		  sigma_hat = sqrt( sigma_hat /
				    (double)(num_pts - muse_params.p()) );

		(*sf_it)->sigma( sigma_hat );
	    } // for
			
	    
	    
	    // Now run through each valid point in the window, assign
	    // that point to the surface to which it has the smallest
	    // residual, normalized by the surface's standard deviation,
	    // if this residual distance falls within the inlier bound
	    for (ii = 0; ii < win.rows(); ii++)
	      for (jj = 0; jj < win.cols(); jj++)
	      {
		px_stat = win[ii][jj];

		// check that range image point is valid
		if ( ri.is_valid( px_stat->row(), px_stat->col() ))
		{
		  // run through each surface
		  for( sf_it = win_surfaces.begin(), opt_d = HUGE_VAL,
			 opt_sf_it = NULL;
		       sf_it != win_surfaces.end(); ++sf_it)
		  {
		    // calculate the residual distance
		    d = fabs(((*sf_it)->surface())
			     ->distance(*(px_stat->position())));

		    // convert to a normalized distance
		    d /= (*sf_it)->sigma();

		    // check if potential inlier
		    if (d < muse_params.mu())         // d already normalized
		      if (d < opt_d) {                // check if best surface
			opt_d = d;
			opt_sf_it = sf_it;
		      }
		  } // for ( sf_it ...

		  // if point can be assigned as an inlier to a suface,
		  // then assign the point to its best fitting surface
		  if (opt_d != HUGE_VAL)   
		  {                            // pt is an inlier to a surface
		    (*opt_sf_it)->points()->push_back(px_stat->position());
		    (*opt_sf_it)->pixels()->push_back(px_stat);
		  } // if (opt_d...
		} // if ( ri.is_valid...
	      } // for (jj...

	    // check each surface to determine if it still maintains enough
	    // data to be considered a nonrandom surface 
	    for ( sf_it = win_surfaces.begin(); sf_it != win_surfaces.end();
		  ++sf_it)
	      if (! ((*SurfaceSurviveHook)(*(*sf_it), muse_params)))
	      {                                      // surface has starved
		cout << "\tKilling a surface" << endl;
		// grab the sf_stat
		sf_stat = *sf_it;

		// calling erase now would invalidate the iterator,
		// so we'll mark the position for later deletion
		*sf_it = NULL;

		// move the pixels associated with this surface to the
		// rescued pixels list.  Later, we will try to reassign
		// these pixels to other surfaces in the window
		copy(sf_stat->pixels()->begin(), sf_stat->pixels()->end(),
		     back_inserter(rescued_pxs));
		
		// free up the SurfaceStat's  memory 
		// (See notes in SurfaceStat.h about memory)
		delete sf_stat->surface();
		delete sf_stat->points();
		delete sf_stat->pixels();
		delete [] (sf_stat->boundingbox());
		delete sf_stat;
	      }
	    
	    // Now delete those surfaces marked from the list.
	    // Note that remove() moves things to the end of the list
	    // and returns an iterator which indicates where the end of the
	    // list would be if the elements had actually been moved.
	    // So we must call erase() after remove() to actually delete
	    // and to set the actual end of list to the correct spot.
	    sf_it = remove(win_surfaces.begin(), win_surfaces.end(),
			   (SurfaceStat *)NULL);
	    win_surfaces.erase(sf_it, win_surfaces.end());


	    // if any surfaces remain, try to reassign rescued pixels to
	    // to the remaining surfaces, then have each surface do a LS
	    // fit to its point set and save surfaces in the global list
	    if (!win_surfaces.empty())
	    {
	      // reassign rescued points
	      for (px_it=rescued_pxs.begin(); px_it!=rescued_pxs.end();px_it++)
	      {
		// grab the pointer to the pixel stat
		px_stat = *px_it;
		
		for( sf_it = win_surfaces.begin(), opt_d = HUGE_VAL,
		       opt_sf_it = NULL;
		     sf_it != win_surfaces.end(); ++sf_it)
		{
		  // calculate the residual distance
		  d = fabs(((*sf_it)->surface())
			   ->distance(*(px_stat->position())));

		  // convert to a normalized distance
		  d /= (*sf_it)->sigma();

		  // check if potential inlier
		  if (d < muse_params.mu())         // d already normalized
		    if (d < opt_d) {                // check if best surface
		      opt_d = d;
		      opt_sf_it = sf_it;
		    }
		} // for ( sf_it ...

		// if point can be assigned as an inlier to a suface,
		// then assign the point to its best fitting surface
		if (opt_d != HUGE_VAL)   
		{                            // pt is an inlier to a surface
		  (*opt_sf_it)->points()->push_back(px_stat->position());
		  (*opt_sf_it)->pixels()->push_back(px_stat);
		} // if (opt_d...
	      }
	      
              // perform LS fit to inliers of each surface,
	      for_each(win_surfaces.begin(), win_surfaces.end(), 
			 LSFitSurfaceStatPtr());

	      // for each surface in the window, mark its inliers in the
	      // segment image, project its inliers to the surface,
	      // compute its bounding box, and release the points and
	      // pixel storage
	      //
	      // Again, overwriting ri with projected data and using the
	      // segment image precludes the use of overlapping windows.
	      // However, it does make the memory constraints manageable
	      // (approximately 16 MB on a 512x512 xyz range image).
	      for( sf_it = win_surfaces.begin(), ii=all_surfaces.size()+1;
		   sf_it != win_surfaces.end(); ++sf_it, ++ii)
	      {
		for (px_it  = (*sf_it)->pixels()->begin();
		     px_it != (*sf_it)->pixels()->end(); ++px_it)
		{
		  // set the segment identifier
		  segment[(*px_it)->row()*ri.cols() + (*px_it)->col()] = ii;

		  // project the pixel onto the surface
		  point=(*sf_it)->surface()->project(*((*px_it)->position()));
		  *(ri[(*px_it)->row()][(*px_it)->col()]) = point;
		}

		// compute the bounding box of the projected points,
		// Note SurfaceStat stores a pointer to a vector of pointers
		// to points. By projecting the points above, the pointers in
		// the above vector already point to reconstructed points.
		ComputeBoundingbox((*sf_it)->boundingbox(),(*sf_it)->points());

		// release the memory allocated for the pixels and points,
		// be careful not to re-release this memory later
		delete (*sf_it)->points(); 
		delete (*sf_it)->pixels();
		(*sf_it)->points( (vector<const Vector3*> *) NULL );
		(*sf_it)->pixels( (vector<PixelStat*> *) NULL );
	      }
	    }
	    
	    // add the window's surfaces to global list, note: back_inserter
	    copy(win_surfaces.begin(), win_surfaces.end(),
		 back_inserter(all_surfaces));

	} // for j

    cout << endl << endl << "Finished scanning... ";
    cout << "extracted " << all_surfaces.size() << " surfaces." << endl;

    // write out the output range image...
    //
    // Mark all pixels not assigned to a surface as invalid.
    cout << "Setting up output range image." << endl;
    for (i = 0; i < ri.rows(); i++)
      for (j = 0; j < ri.cols(); j++)
	if (segment[i*ri.cols() + j] == 0)
	  ri.mark_invalid(i, j);

    // Now output the range image
    ri.write( argv[3] );
    

    // prepare the dx file to receive the vertices for the bounding rectangles
    cout << "Outputting surface file in IBM DataExplorer format." << endl;
    WriteDXHeader(surfacefile, all_surfaces);

    // output the vertices for the bounding boxes
    for (sf_it = all_surfaces.begin(); sf_it != all_surfaces.end(); sf_it++)
      WriteDXSurfaceVertices(surfacefile, (*sf_it)->boundingbox());

    // output the remainder of the dx file...
    WriteDXSurfaceDataConnectionsFields(surfacefile, all_surfaces);
    surfacefile.close();

    
    // output the segmentation file if necessary,
    //     Segmentation format is a binary image of unsigned integers.  The
    //     first two integers are the number of rows and columns.  Then
    //     rows*columns unsigned integers specifying a surface number.
    //     Surfaces start counting at 1.  A value of 0 indicates that the pixel
    //     is not labelled.
    //
    // IMPORTANT:  With this current format, a pixel can only be described
    //     by one surface.  When we move to overlapping windows, this will
    //     become a problem.
    //
    if (output_segment) {
      cout << "Writing segmentation file " << argv[5] << endl;
      int ROWS = ri.rows(), COLS = ri.cols();
      segmentfile.write( (char *)&ROWS, sizeof(int))
	         .write( (char *)&COLS, sizeof(int))
	         .write( (char *)segment, ROWS*COLS*sizeof(unsigned int));
      segmentfile.close();
      delete [] segment;
    }
	
    // Clean up.
    //
    // run through the global surface list and free up the memory for
    // each surface and boundingbox.  Remember that the point and pixel
    // lists have already been freed.
    //
    // See notes in SurfaceStat.h
    cout << "Cleaning up." << endl;
    for (sf_it = all_surfaces.begin(); sf_it != all_surfaces.end(); ++sf_it)
    {
	delete (*sf_it)->surface();
	delete [] ((*sf_it)->boundingbox());
    }
    delete [] segment;


    exit(0);		    
}




//
// SimpleDone returns true if the number of available datapoints in a window
// falls below a threshold
//
// win is passed to any WindowDoneHook routine, so that more elegant
// methods based on the actual remaining data points can be used (MINPRAN)
//
// The parameter num_surfaces is the number of surfaces already extracted
// from this window.
//
// All WindowDoneHooks should make sure that numavail > k0 + p
//
bool SimpleDone(WindowStat<PixelStat> &win, MuseParameters &mp,
		int num_surfaces, int k0, int numavail)
{
    return ((numavail <= MY_MAX((mp.p() + k0), mp.min_feature()))
	     || (num_surfaces >= mp.extract_number()));
}


//
// A naive random sampling routine
//
// A list of random sample points are returned in ran_samples.
// Only the "available" points in win are allowed to be random sample points.
//
// Routine clears the random sample vector prior to filling.
// Routine assumes that one can actually succeed at finding p random sample
// points. If the number of available points falls below p, then infinite loop.
//
bool SimpleSample(WindowStat<PixelStat> &win, int p,
		  vector<PixelStat*> &ran_samples)
{
    int i, r, c;
    bool done;
    PixelStat* px_stat;
    vector<PixelStat*>::iterator where;

    assert( p > 0 );

    // clear the return vector
    if (!ran_samples.empty())
	ran_samples.erase( ran_samples.begin(), ran_samples.end() );

    // generate the random samples
    for (i=0; i < p; i++)
    {
	done = false;

	while (!done)
	{
	    // grab a random pixel in the window
	    r = rand() % win.rows();
	    c = rand() % win.cols();
	    px_stat = win[r][c];

	    // check to make sure pixel is available
	    if (px_stat->available())
	    {
		// check to make sure pixel has not already been selected
		// as a random sample point
		where = find( ran_samples.begin(), ran_samples.end(), px_stat);

		if (where == ran_samples.end()) // not currently in the list
		{
		    // add it to the list
		    ran_samples.push_back( px_stat ); 

		    done = true;   // we have found one sample point
		} // if
	    } // if
	} // while
    } // for

    return ( ran_samples.size() == p );
}



//
// Gather all the points in window that are available AND are NOT
// the random sample points
//
// This routine empties the pts vector prior to filling
//
bool ScanWindow(WindowStat<PixelStat> &win, vector<PixelStat*> &ran_samples,
		vector<PixelStat*> &pts)
{
    int i, j;
    PixelStat *px_stat;
    vector<PixelStat*>::iterator where;
    

    // clear the return vector
    if (!pts.empty())
	pts.erase( pts.begin(), pts.end() );

    for (i=0; i < win.rows(); i++)
	for (j=0; j < win.cols(); j++)
	{
	    // grab the PixelStat*
	    px_stat = win[i][j];

	    if (px_stat->available())
	    {
		// check if point is a random sample point
		where = find(ran_samples.begin(), ran_samples.end(), px_stat);

		if (where == ran_samples.end()) // not in the ran_sample list
		    pts.push_back( px_stat );  // store as pt to return
	    }
	}


    return ( !pts.empty() );
}



//
// Scan the "interior" of the random sample and return the available
// "interior" points.
//
// Feasible assumptions:  interior is defined as interior of the convex hull
//                        of the random sample points
//
// Again, we'd like to NOT return the random sample points
//
bool ScanSample(WindowStat<PixelStat> &win, vector<PixelStat*> &ran_samples,
		vector<PixelStat*> &pts)
{

    //  For now, we'll concentrate on scanning triangles, testing
    //  each point in the window for whether it lies "inside" the triangle.
    //  We'll come back later and put in a real convex set scan converter
    int i, j;
    PixelStat *px_stat;
    vector<PixelStat*>::iterator where;

    if (ran_samples.size() != 3)
    {
	cerr << "ScanSample only supports scan converting triangles. "
	     << "Exiting." << endl;
	exit(0);
    }

    // clear the return vector
    if (!pts.empty())
	pts.erase( pts.begin(), pts.end() );

    //
    // Our approach right now is to consider an angle of the triangle
    // as the origin of a coordinate system with basis vectors given
    // by the edges "out of" that vertex.  If a point is in the
    // "first quadrant" of the coordinate system from the first vertex
    // and is in the "first quadrant" of the coordinate system of the
    // second vertex, then the point is in the triangle.
    //
    // a vector v in the xy basis can be written as
    //        [v]xy = a v1 + b v2
    // where [a b] is the vector v in the (v1, v2) basis
    //
    // therefore [v]xy = [v1 v2][a b]'
    // and       [a b]' = [v1 v2]^{-1}[v]xy
    //
    // This method should be replaced with an actual scan converter.
    // Although the code would be more complex, a graphics scane converter
    // will gather the points more accurately AND be faster.
    //
    double v11[2],   // 1st basis vector from first vertex
	   v12[2],   // 2nd basis vector from first vertex
	   v21[2],   // 1st basis vector from second vertex
	   v22[2],   // 2nd basis vector from second vertex
	   v1xy[2],  // vector from 1st vertex to the point in question
  	   v2xy[2];  // vector from 2nd vertex to the point in question
    double a1, b1, a2, b2, det;
    double v1_inv[2][2], v2_inv[2][2];
    
    // calculate the basis vectors
    v11[0] = ran_samples[1]->col() - ran_samples[0]->col();
    v11[1] = ran_samples[1]->row() - ran_samples[0]->row();
    v12[0] = ran_samples[2]->col() - ran_samples[0]->col();
    v12[1] = ran_samples[2]->row() - ran_samples[0]->row();

    v21[0] = ran_samples[0]->col() - ran_samples[1]->col();
    v21[1] = ran_samples[0]->row() - ran_samples[1]->row();
    v22[0] = ran_samples[2]->col() - ran_samples[1]->col();
    v22[1] = ran_samples[2]->row() - ran_samples[1]->row();

    // compute inverse basis transformation
    det = v11[0]*v12[1] - v12[0]*v11[1];
    v1_inv[0][0] =  v12[1]/det;
    v1_inv[0][1] = -v12[0]/det;
    v1_inv[1][0] = -v11[1]/det;
    v1_inv[1][1] =  v11[0]/det;
    if (det == 0.0)
	return false;
	
    
    det = v21[0]*v22[1] - v22[0]*v21[1];
    v2_inv[0][0] =  v22[1]/det;
    v2_inv[0][1] = -v22[0]/det;
    v2_inv[1][0] = -v21[1]/det;
    v2_inv[1][1] =  v21[0]/det;
    if (det == 0.0)
	return false;

    
    for (i=0; i < win.rows(); i++)
	for (j=0; j < win.cols(); j++)
	{
	    // grab the PixelStat*
	    px_stat = win[i][j];

	    if (px_stat->available())
	    {
		// check if point is a random sample point
		where = find(ran_samples.begin(), ran_samples.end(), px_stat);

		if (where == ran_samples.end()) // not in the ran_sample list
		{
		    // calculate the vectors from the point in question
		    // to vertex 1 and vertex 2
		    v1xy[0] = px_stat->col() - ran_samples[0]->col();
		    v1xy[1] = px_stat->row() - ran_samples[0]->row();

    		    v2xy[0] = px_stat->col() - ran_samples[1]->col();
		    v2xy[1] = px_stat->row() - ran_samples[1]->row();

		    // check if point is in the 1st quadrant of vertex 1's
		    // basis AND in the 1st quadrant of vertex 2's basis
		    a1 = v1_inv[0][0]*v1xy[0] + v1_inv[0][1]*v1xy[1];
		    b1 = v1_inv[1][0]*v1xy[0] + v1_inv[1][1]*v1xy[1];

		    a2 = v2_inv[0][0]*v2xy[0] + v2_inv[0][1]*v2xy[1];
		    b2 = v2_inv[1][0]*v2xy[0] + v2_inv[1][1]*v2xy[1];

		    if ((a1 >= 0.0) && (b1 >= 0.0)
			&& (a2 >= 0.0) && (b2 >= 0.0))
			pts.push_back( px_stat );  // store the pt 
		}
	    }
	}

#if 0
    for (i=0; i < win.rows(); i++)
    {
	for (j=0; j < win.cols(); j++)
	{
	    px_stat = win[i][j];
	    if (!px_it->available())
	    {	cout << "x";  continue;    }

	    where = find(pts.begin(), pts.end(), px_stat);
	    if (where != pts.end())
	    {	cout << "o";  continue;    }
	    
	    px_stat = win[i][j];
	    where = find(ran_samples.begin(), ran_samples.end(), px_stat);
	    if (where != ran_samples.end())
	    {	cout << "*";  continue;    }
	    cout << ".";
	}
	cout << endl;
    }
    cout << endl << endl;
#endif

    
    return ( !pts.empty() );
}



//
// How many points in the window are still available?
//
int  NumAvail(WindowStat<PixelStat> &win)
{
    int i,j, count=0;

    for (i=0; i < win.rows(); i++)
	for (j=0; j < win.cols(); j++)
	    if (win[i][j]->available())
		count++;

    return count;
}


//
// Create the surface defined by the random sample
//
Surface* DefineRandomSampleFit(vector<PixelStat*>& ran_sample)
{
    Surface *surf = NULL;
    
    switch (ran_sample.size())
    {
      case 3:    // plane
	// create a new plane, defined by the random sample
	surf = new Plane(*(ran_sample[0]->position()),
			 *(ran_sample[1]->position()),
			 *(ran_sample[2]->position()));

	break;

      default:   // undefined fitting primitive
	cerr << "Error: the only supported primitive is "
	     << "currently planes (p=3)." << endl;
	exit(0);
    }

    return surf;
}



//
// Determine if a Surface can be considered non-random.  This simple version
// simple checks to make sure there are enough inliers.  A more sophisticated
// routine may look at the actual points to see if there is any non-random
// structure in the points.
//
// Also checking how a simple kill of surfaces with a std dev > 1 cm does.
//   (Removed since Perceptron images can have a std dev > 1cm)
//
bool SimpleSurfaceSurvive(SurfaceStat& sf_stat, MuseParameters &mp)
{
    return ( (sf_stat.points()->size() > mp.min_feature()) );
}


//
// Find the indices for the middle and maximum standard deviation (out of 3)
//
// W holds 3 values
//
void FindMidMaxStdDev(double W[], int &i_mid, int &i_max)
{
	if ((W[1] >= W[2]) && (W[1] >= W[3]))
	    i_max = 1;
	else if ((W[2] >= W[1]) && (W[2] >= W[3]))
	    i_max = 2;
	else if ((W[3] >= W[1]) && (W[3] >= W[2]))
	    i_max = 3;
	
	if (((W[2] >= W[1]) && (W[1] >= W[3])) ||
	    ((W[3] >= W[1]) && (W[1] >= W[2])))
	    i_mid = 1;
	else if (((W[1] >= W[2]) && (W[2] >= W[3])) ||
		 ((W[3] >= W[2]) && (W[2] >= W[1])))
	    i_mid = 2;
	else if (((W[1] >= W[3]) && (W[3] >= W[2])) ||
		 ((W[2] >= W[3]) && (W[3] >= W[1])))
	    i_mid = 3;
}


//
// Routine to write the DX header information and prepare the file to
// for receiving the vertices
//
void WriteDXHeader(ofstream &surfacefile,
		   vector<SurfaceStat*> &all_surfaces)
{
    // start writing out the surface output file
    surfacefile << "# This dx file contains the surfaces extracted by muse."
		<< endl
		<< "# The surfaces appear as a field where the data "
		<< "for each surface is" << endl
		<< "# the estimated sigma" << endl << endl;

    surfacefile << "# Vertices of each bounding rectangle" << endl;
    surfacefile << "object \"vertices\" class array type float rank 1 shape 3 "
		<< "items " << 4*all_surfaces.size() << " data follows"
		<< endl;
    // the next thing written to the surfacefile should be the vertices
    // of all the bounding rectangles
}


//
// Routine to write the bounding rectangle vertices into the DX vertex list
//
// Note "corner" is assumed to be a vector containing 4 Vector3's, containing
// the 3D locations of the corners of a rectangle, the corners are "numbered"
//          3--------2
//          |        |
//          |        |
//          0--------1
//
// but for data explorer the vertices are "outputted" in order 0,1,3,2
//
void WriteDXSurfaceVertices(ofstream &surfacefile, Vector3 corner[])
{
    // need to make sure all "surfaces" have a consistent ordering
    // so that all surface normals are "front facing"
    //
    // Also note that our camera files are left-handed, so our
    // extracted surfaces will also be left-handed.
    if (((corner[1] - corner[0]) % (corner[3] - corner[0]))
	* Vector3(0, 0, -1) >= 0.0)
    {
	surfacefile << corner[0].x() << " " << corner[0].y() << " "
		    << corner[0].z() << endl;
	surfacefile << corner[1].x() << " " << corner[1].y() << " "
		    << corner[1].z() << endl;
	surfacefile << corner[3].x() << " " << corner[3].y() << " "
		    << corner[3].z() << endl;
	surfacefile << corner[2].x() << " " << corner[2].y() << " "
		    << corner[2].z() << endl;
    }
    else
    {
	surfacefile << corner[0].x() << " " << corner[0].y() << " "
		    << corner[0].z() << endl;
	surfacefile << corner[3].x() << " " << corner[3].y() << " "
		    << corner[3].z() << endl;
	surfacefile << corner[1].x() << " " << corner[1].y() << " "
		    << corner[1].z() << endl;
	surfacefile << corner[2].x() << " " << corner[2].y() << " "
		    << corner[2].z() << endl;
    }
}



//
// Routine to write some of the components of the DataExplorer surface file.
// This routine should only be called after the header and vertices
// have been written to the data explorer file
//
void WriteDXSurfaceDataConnectionsFields(ofstream &surfacefile,
					 vector<SurfaceStat*> &all_surfaces)
{
    int i;
    vector<SurfaceStat*>::iterator sf_it;

    // Now we'll write out the data for the surfaces
    surfacefile << endl << endl
		<< "# Data is the std dev for each fit." << endl;
    surfacefile << "object \"std dev\" class array type float rank 1 shape 1 "
		<< "items " << all_surfaces.size() << " data follows"
		<< endl;
    for (sf_it = all_surfaces.begin(), i=1; sf_it != all_surfaces.end();
	 ++sf_it, ++i)
    {
	surfacefile << " " << (*sf_it)->sigma();
	if (i % 5 == 0) surfacefile << endl;
    }
    surfacefile << endl
		<< "attribute \"dep\" string \"connections\" " << endl;


    // we still need to write out the connections, note the special
    // ordering of the vertices, we wrote out the positions or vertices
    // of each bounding rectangle as 0, 1, 3, 2
    surfacefile << endl << endl
		<< "# Bounding rectangle connections" << endl;
    surfacefile << "object \"bounding rectangles\" class array type int "
		<< "rank 1 shape 4 items " << all_surfaces.size()
		<< " data follows" << endl;
    for (i=0; i < 4*all_surfaces.size(); i+=4)
	surfacefile << i << " " << i+1 << " " << i+2 << " " << i+3 << endl;
    surfacefile << "attribute \"element type\" string \"quads\" " << endl
		<< "attribute \"ref\" string \"positions\" " << endl;

    
    // now write out the field
    surfacefile << endl << endl
		<< "# Object containing the std dev field" << endl;
    surfacefile << "object \"surfaces with std dev\" class field" << endl
		<< "component \"positions\" value \"vertices\" " << endl
		<< "component \"connections\" value \"bounding rectangles\" "
		<< endl
		<< "component \"data\" value \"std dev\" " << endl << endl;

    // set the default surface
    surfacefile << "# setting the default object" << endl
		<< "default \"surfaces with std dev\" " << endl << endl
		<< "end" << endl << endl;    
}



//
// Compute the bounding box for the a set of reconstructed points.
// This routine currently only computes a bounding rectangle of the
// points projected onto planes.  Therefore we assume that boundingbox
// contains 4 Vector3's and that all the points fall on a plane.
//
// The bounding box is computed by finding the principal components of
// the projected points.  Since all the points are on a plane, one
// of the pc's should be zero (or at least very very small).  The eigenvectors
// for the two larger pc's provide an orthogonal coordinate system that
// spans the projected data.  The points in this coordinate system are
// then searched for min/max values.
//
//
// The bounding box is stored in the vector boundingbox.  Note the ordering
// of the vertices
//          3--------2
//          |        |
//          |        |
//          0--------1
//
void ComputeBoundingbox( Vector3 *boundingbox, vector<const Vector3*> *points)
{
    assert( (boundingbox != NULL) && (points != NULL));

    vector<const Vector3*>::iterator pt_cit;
    Vector3 pt_mean, pt;
    Vector3 v1, v2;
    double **X, *W, **V;
    int i_mid, i_max;   // indexes for the largest and middle eigenvalue
    unsigned int i;
    
    // allocate the matrices for SVD
    V = dmatrix(1, 3, 1, 3);  // Note we are using Numerical Recipes matrices
    W = dvector(1, 3);
    X = dmatrix(1, points->size(), 1, 3);
    pt_mean.x(0.0).y(0.0).z(0.0);

    // build the "data" matrix
    for (pt_cit = points->begin(), i=1; pt_cit != points->end(); pt_cit++, i++)
    {
      pt = **pt_cit;
      X[i][1] = pt.x();  X[i][2] = pt.y();  X[i][3] = pt.z();
      pt_mean += pt;
    }

    // center the data
    pt_mean /= (double)points->size();
    for (i=1; i <= points->size(); i++) {
      X[i][1] -= pt_mean.x();  X[i][2] -= pt_mean.y(); X[i][3] -= pt_mean.z();
    }
    
    // compute the PC's of the projected points
    svdcmp(X, points->size(), 3, W, V);

    // W holds sqrt's of eigenvalues of X'X and
    // V holds eigenvectors of X'X

    // find the largest and next to large std deviations
    FindMidMaxStdDev(W, i_mid, i_max);

    // assign the direction vectors
    v1.x(V[1][i_max]).y(V[2][i_max]).z(V[3][i_max]);
    v2.x(V[1][i_mid]).y(V[2][i_mid]).z(V[3][i_mid]);


    // Compute the corners of the bounding polygon rectangle on the surface
    // This requires searching the data in pc coordinates, which were
    // returned by SVD in the matrix U (which overwrote the matrix X)
    // We'll use v1 and v2 as shorthand for coordinate basis for the
    // points projected onto the plane, with v1 corresponding to the
    // "x-axis" of the points and v2 corresponding to the "y-axis" of
    // the points.
	
    // Note that the corners of the rectangle are "numbered"
    //          3--------2
    //          |        |
    //          |        |
    //          0--------1
    //
    // but for data explorer the vertices are "outputted" in order 0,1,3,2

    double xmin = HUGE_VAL, xmax = -HUGE_VAL,
           ymin = HUGE_VAL, ymax = -HUGE_VAL;
    for (i=1; i <= points->size(); i++)
    {
      xmin = MY_MIN(xmin, X[i][i_max]);
      xmax = MY_MAX(xmax, X[i][i_max]);
      ymin = MY_MIN(ymin, X[i][i_mid]);
      ymax = MY_MAX(ymax, X[i][i_mid]);
    }

    // Note the the pc coordinates U (which overwrote X) are variables
    // standardized to have variance 1/(n-1) (See Jolliffe pg. 38)
    //
    // Thus we need to convert U scores to PC scores with their
    // respective variance. zi = ui * W[i]
    boundingbox[0] = pt_mean + v1*xmin*W[i_max] + v2*ymin*W[i_mid];
    boundingbox[1] = pt_mean + v1*xmax*W[i_max] + v2*ymin*W[i_mid];
    boundingbox[2] = pt_mean + v1*xmax*W[i_max] + v2*ymax*W[i_mid];
    boundingbox[3] = pt_mean + v1*xmin*W[i_max] + v2*ymax*W[i_mid];
		    
    
    free_dmatrix(X, 1, points->size(), 1, 3); 
    free_dmatrix(V, 1, 3, 1, 3);
    free_dvector(W, 1, 3);
}
